from typing import List, Dict, Pattern, Optional
from collections import namedtuple
from decimal import Decimal, getcontext
import copy
import regex

from recognizers_text.utilities import RegExpUtility
from recognizers_text.culture import Culture
from recognizers_text.extractor import ExtractResult
from recognizers_text.parser import ParseResult
from recognizers_number.resources.japanese_numeric import JapaneseNumeric
from recognizers_number.number.cjk_parsers import CJKNumberParserConfiguration
from recognizers_number.number.parsers import BaseNumberParser, NumberParserConfiguration
from recognizers_number.culture import CultureInfo

getcontext().prec = 15


class JapaneseNumberParserConfiguration(CJKNumberParserConfiguration):
    @property
    def cardinal_number_map(self) -> Dict[str, int]:
        return dict()

    @property
    def ordinal_number_map(self) -> Dict[str, int]:
        return dict()

    @property
    def round_number_map(self) -> Dict[str, int]:
        return self._round_number_map

    @property
    def culture_info(self) -> CultureInfo:
        return self._culture_info

    @property
    def digital_number_regex(self) -> Pattern:
        return self._digital_number_regex

    @property
    def fraction_marker_token(self) -> str:
        return self._fraction_marker_token

    @property
    def negative_number_sign_regex(self) -> Pattern:
        return self._negative_number_sign_regex

    @property
    def half_a_dozen_regex(self) -> Pattern:
        return None

    @property
    def half_a_dozen_text(self) -> str:
        return self._half_a_dozen_text

    @property
    def lang_marker(self) -> str:
        return self._lang_marker

    @property
    def non_decimal_separator_char(self) -> str:
        return self._non_decimal_separator_char

    @property
    def decimal_separator_char(self) -> str:
        return self._decimal_separator_char

    @property
    def word_separator_token(self) -> str:
        return self._word_separator_token

    @property
    def zero_char(self) -> str:
        return self._zero_char

    @property
    def pair_char(self) -> str:
        return self._pair_char

    @property
    def written_decimal_separator_texts(self) -> List[str]:
        return list()

    @property
    def written_group_separator_texts(self) -> List[str]:
        return list()

    @property
    def written_integer_separator_texts(self) -> List[str]:
        return list()

    @property
    def written_fraction_separator_texts(self) -> List[str]:
        return list()

    @property
    def zero_to_nine_map(self) -> Dict[str, int]:
        return self._zero_to_nine_map

    @property
    def round_number_map_char(self) -> Dict[str, int]:
        return self._round_number_map_char

    @property
    def full_to_half_map(self) -> Dict[str, int]:
        return self._full_to_half_map

    @property
    def trato_sim_map(self) -> Dict[str, int]:
        return None

    @property
    def unit_map(self) -> Dict[str, int]:
        return self._unit_map

    @property
    def round_direct_list(self) -> List[str]:
        return self._round_direct_list

    @property
    def ten_chars(self) -> List[str]:
        return self._ten_chars

    @property
    def digit_num_regex(self) -> Pattern:
        return self._digit_num_regex

    @property
    def dozen_regex(self) -> Pattern:
        return self._dozen_regex

    @property
    def percentage_regex(self) -> Pattern:
        return self._percentage_regex

    @property
    def double_and_round_regex(self) -> Pattern:
        return self._double_and_round_regex

    @property
    def frac_split_regex(self) -> Pattern:
        return self._frac_split_regex

    @property
    def point_regex(self) -> Pattern:
        return self._point_regex

    @property
    def spe_get_number_regex(self) -> Pattern:
        return self._spe_get_number_regex

    @property
    def pair_regex(self) -> Pattern:
        return self._pair_regex

    @property
    def round_number_integer_regex(self) -> Pattern:
        return self._round_number_integer_regex

    def __init__(self, culture_info=None):
        if culture_info is None:
            culture_info = CultureInfo(Culture.Japanese)

        self._culture_info = culture_info

        self._lang_marker = JapaneseNumeric.LangMarker
        self._decimal_separator_char = JapaneseNumeric.DecimalSeparatorChar
        self._fraction_marker_token = JapaneseNumeric.FractionMarkerToken
        self._non_decimal_separator_char = JapaneseNumeric.NonDecimalSeparatorChar
        self._half_a_dozen_text = JapaneseNumeric.HalfADozenText
        self._word_separator_token = JapaneseNumeric.WordSeparatorToken
        self._zero_char = JapaneseNumeric.ZeroChar
        self._pair_char = JapaneseNumeric.PairChar

        self._round_number_map = JapaneseNumeric.RoundNumberMap
        self._digital_number_regex = RegExpUtility.get_safe_reg_exp(
            JapaneseNumeric.DigitalNumberRegex)

        self._zero_to_nine_map = JapaneseNumeric.ZeroToNineMap
        self._round_number_map_char = JapaneseNumeric.RoundNumberMapChar
        self._full_to_half_map = JapaneseNumeric.FullToHalfMap
        self._unit_map = JapaneseNumeric.UnitMap
        self._round_direct_list = JapaneseNumeric.RoundDirectList
        self._ten_chars = JapaneseNumeric.TenChars
        self._digit_num_regex = JapaneseNumeric.DigitNumRegex
        self._dozen_regex = JapaneseNumeric.DozenRegex
        self._percentage_regex = JapaneseNumeric.PercentageRegex
        self._double_and_round_regex = RegExpUtility.get_safe_reg_exp(
            JapaneseNumeric.DoubleAndRoundRegex)
        self._frac_split_regex = RegExpUtility.get_safe_reg_exp(
            JapaneseNumeric.FracSplitRegex)
        self._negative_number_sign_regex = RegExpUtility.get_safe_reg_exp(
            JapaneseNumeric.NegativeNumberSignRegex)
        self._point_regex = JapaneseNumeric.PointRegex
        self._spe_get_number_regex = RegExpUtility.get_safe_reg_exp(
            JapaneseNumeric.SpeGetNumberRegex)
        self._pair_regex = RegExpUtility.get_safe_reg_exp(
            JapaneseNumeric.PairRegex)
        self._round_number_integer_regex = RegExpUtility.get_safe_reg_exp(
            JapaneseNumeric.RoundNumberIntegerRegex)

    def normalize_token_set(self, tokens: List[str], context: ParseResult) -> List[str]:
        return tokens

    def resolve_composite_number(self, number_str: str) -> int:
        return 0



from typing import List
from enum import Enum

from recognizers_number.number.extractors import ReVal, BaseNumberExtractor
from recognizers_text.utilities import RegExpUtility
from recognizers_number.number.constants import Constants
from recognizers_number.resources.japanese_numeric import JapaneseNumeric


class JapaneseNumberExtractorMode(Enum):
    DEFAULT = 0
    EXTRACT_ALL = 1


class JapaneseNumberExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[ReVal]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM

    def __init__(self, mode: JapaneseNumberExtractorMode = JapaneseNumberExtractorMode.DEFAULT):
        self.__regexes: List[ReVal] = list()

        cardinal_ex = JapaneseCardinalExtractor(mode)
        self.__regexes.extend(cardinal_ex.regexes)

        fraction_ex = JapaneseFractionExtractor()
        self.__regexes.extend(fraction_ex.regexes)


class JapaneseCardinalExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[ReVal]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_CARDINAL

    def __init__(self, mode: JapaneseNumberExtractorMode = JapaneseNumberExtractorMode.DEFAULT):
        self.__regexes: List[ReVal] = list()

        integer_ex = JapaneseIntegerExtractor(mode)
        self.__regexes.extend(integer_ex.regexes)

        double_ex = JapaneseDoubleExtractor()
        self.__regexes.extend(double_ex.regexes)


class JapaneseIntegerExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[ReVal]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_INTEGER

    def __init__(self, mode: JapaneseNumberExtractorMode = JapaneseNumberExtractorMode.DEFAULT):
        self.__regexes = [
            # 123456,  －１２３４５６
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.NumbersSpecialsChars),
                val='IntegerNum'),
            # 15k,  16 G
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.NumbersSpecialsCharsWithSuffix),
                val='IntegerNum'),
            # 1,234,  ２，３３２，１１１
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.DottedNumbersSpecialsChar),
                val='IntegerNum'),
            # 半百  半ダース
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.NumbersWithHalfDozen),
                val='IntegerJpn'),
            # 一ダース  五十ダース
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.NumbersWithDozen),
                val='IntegerJpn')
        ]
        if mode == JapaneseNumberExtractorMode.DEFAULT:
            self.__regexes.append(
                # 一百五十五, 负一亿三百二十二. Uses an allow list to avoid extracting "西九条" from "九"
                ReVal(
                    re=RegExpUtility.get_safe_reg_exp(
                        JapaneseNumeric.NumbersWithAllowListRegex),
                    val='IntegerJpn'
                )
            )
        elif mode == JapaneseNumberExtractorMode.EXTRACT_ALL:
            self.__regexes.append(
                # 一百五十五, 负一亿三百二十二, "西九条" from "九". Uses no allow lists and extracts all potential integers (useful in Units, for example).
                ReVal(
                    re=RegExpUtility.get_safe_reg_exp(
                        JapaneseNumeric.NumbersAggressiveRegex),
                    val='IntegerJpn'
                )
            )


class JapaneseDoubleExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[ReVal]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_DOUBLE

    def __init__(self):
        self.__regexes = [
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.DoubleSpecialsChars),
                val='DoubleNum'),
            # (-)2.5, can avoid cases like ip address xx.xx.xx.xx
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.DoubleSpecialsCharsWithNegatives),
                val='DoubleNum'),
            # (-).2
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.SimpleDoubleSpecialsChars),
                val='DoubleNum'),
            # 1.0 K
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.DoubleWithMultiplierRegex),
                val='DoubleNum'),
            # １５.２万
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.DoubleWithThousandsRegex),
                val='DoubleJpn'),
            # 2e6, 21.2e0
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.DoubleExponentialNotationRegex),
                val='DoublePow'),
            # 2^5
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.DoubleScientificNotationRegex),
                val='DoublePow')
        ]


class JapaneseFractionExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[ReVal]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_FRACTION

    def __init__(self):
        self.__regexes = [
            # -4 5/2,  ４ ６／３
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.FractionNotationSpecialsCharsRegex),
                val='FracNum'),
            # 8/3
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.FractionNotationRegex),
                val='FracNum'),
            # 五分の二 七分の三
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.AllFractionNumber),
                val='FracJpn')
        ]


class JapaneseOrdinalExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[ReVal]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_ORDINAL

    def __init__(self):
        self.__regexes = [
            # だい一百五十四
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.OrdinalRegex),
                val='OrdinalJpn'),
            # だい２５６５
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.OrdinalNumbersRegex),
                val='OrdinalJpn'),
            # 2折 ２.５折
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.NumbersFoldsPercentageRegex),
                val='OrdinalJpn')
        ]


class JapanesePercentageExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[ReVal]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_PERCENTAGE

    def __init__(self):
        self.__regexes = [
            # 百パーセント 十五パーセント
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.SimplePercentageRegex),
                val='PerJpn'),
            # 19パーセント　１パーセント
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.NumbersPercentagePointRegex),
                val='PerNum'),
            # 3,000パーセント  １，１２３パーセント
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.NumbersPercentageWithSeparatorRegex),
                val='PerNum'),
            # 3.2 k パーセント
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.NumbersPercentageWithMultiplierRegex),
                val='PerNum'),
            # 15kパーセント
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.SimpleNumbersPercentageWithMultiplierRegex),
                val='PerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.SimpleIntegerPercentageRegex),
                val='PerNum'),
            # 2割引 ２.５割引
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.NumbersFoldsPercentageRegex),
                val='PerSpe'),
            # 三割引 六点五折 七五折
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.FoldsPercentageRegex),
                val='PerSpe'),
            # 5割 7割半
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.SimpleFoldsPercentageRegex),
                val='PerSpe'),
            # 七割半
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.SpecialsPercentageRegex),
                val='PerSpe'),
            # 2割 ２.５割
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.NumbersSpecialsPercentageRegex),
                val='PerSpe'),
            # 三割
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.SimpleSpecialsPercentageRegex),
                val='PerSpe'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    JapaneseNumeric.SpecialsFoldsPercentageRegex),
                val='PerSpe')
        ]



from typing import Pattern, List, NamedTuple
import regex

from recognizers_text.utilities import RegExpUtility
from recognizers_number.number.models import NumberMode, LongFormatMode
from recognizers_number.resources import BaseNumbers
from recognizers_number.resources.french_numeric import FrenchNumeric
from recognizers_number.number.extractors import ReVal, ReRe, BaseNumberExtractor, BasePercentageExtractor
from recognizers_number.number.constants import Constants


class FrenchNumberExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[ReVal]:
        return self.__regexes

    @property
    def ambiguity_filters_dict(self) -> List[ReRe]:
        return self.__ambiguity_filters_dict

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM

    @property
    def _negative_number_terms(self) -> Pattern:
        return self.__negative_number_terms

    def __init__(self, mode: NumberMode = NumberMode.DEFAULT):
        self.__negative_number_terms = RegExpUtility.get_safe_reg_exp(
            FrenchNumeric.NegativeNumberTermsRegex)
        self.__regexes: List[ReVal] = list()
        cardinal_ex: FrenchCardinalExtractor = None

        if mode is NumberMode.PURE_NUMBER:
            cardinal_ex = FrenchCardinalExtractor(
                FrenchNumeric.PlaceHolderPureNumber)
        elif mode is NumberMode.CURRENCY:
            self.__regexes.append(ReVal(re=RegExpUtility.get_safe_reg_exp(
                FrenchNumeric.CurrencyRegex), val='IntegerNum'))

        if cardinal_ex is None:
            cardinal_ex = FrenchCardinalExtractor()

        self.__regexes.extend(cardinal_ex.regexes)

        fraction_ex = FrenchFractionExtractor(mode)
        self.__regexes.extend(fraction_ex.regexes)

        ambiguity_filters_dict: List[ReRe] = list()

        if mode != NumberMode.Unit:
            for key, value in FrenchNumeric.AmbiguityFiltersDict.items():
                ambiguity_filters_dict.append(ReRe(reKey=RegExpUtility.get_safe_reg_exp(key),
                                                   reVal=RegExpUtility.get_safe_reg_exp(value)))
        self.__ambiguity_filters_dict = ambiguity_filters_dict


class FrenchCardinalExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[ReVal]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_CARDINAL

    def __init__(self, placeholder: str = FrenchNumeric.PlaceHolderDefault):
        self.__regexes: List[ReVal] = list()

        # Add integer regexes
        integer_ex = FrenchIntegerExtractor(placeholder)
        self.__regexes.extend(integer_ex.regexes)

        # Add double regexes
        double_ex = FrenchDoubleExtractor(placeholder)
        self.__regexes.extend(double_ex.regexes)


class FrenchIntegerExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[NamedTuple('re_val', [('re', Pattern), ('val', str)])]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_INTEGER

    def __init__(self, placeholder: str = FrenchNumeric.PlaceHolderDefault):
        self.__regexes = [
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    FrenchNumeric.NumbersWithPlaceHolder(placeholder), regex.I),
                val='IntegerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    FrenchNumeric.NumbersWithSuffix, regex.S),
                val='IntegerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(self._generate_format_regex(
                    LongFormatMode.INTEGER_DOT, placeholder), regex.V1),
                val='IntegerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(self._generate_format_regex(
                    LongFormatMode.INTEGER_BLANK, placeholder)),
                val='IntegerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(self._generate_format_regex(
                    LongFormatMode.INTEGER_NO_BREAK_SPACE, placeholder)),
                val='IntegerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    FrenchNumeric.RoundNumberIntegerRegexWithLocks),
                val='IntegerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    FrenchNumeric.NumbersWithDozenSuffix),
                val='IntegerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    FrenchNumeric.AllIntRegexWithLocks),
                val='IntegerFr'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    FrenchNumeric.AllIntRegexWithDozenSuffixLocks),
                val='IntegerFr')
        ]


class FrenchDoubleExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[NamedTuple('re_val', [('re', Pattern), ('val', str)])]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_DOUBLE

    def __init__(self, placeholder):
        self.__regexes = [
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    FrenchNumeric.DoubleDecimalPointRegex(placeholder)),
                val='DoubleNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    FrenchNumeric.DoubleWithoutIntegralRegex(placeholder)),
                val='DoubleNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(self._generate_format_regex(
                    LongFormatMode.DOUBLE_DOT_COMMA, placeholder)),
                val='DoubleNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(self._generate_format_regex(
                    LongFormatMode.DOUBLE_NO_BREAK_SPACE_COMMA, placeholder)),
                val='DoubleNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    FrenchNumeric.DoubleWithMultiplierRegex),
                val='DoubleNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    FrenchNumeric.DoubleWithRoundNumber),
                val='DoubleNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    FrenchNumeric.DoubleAllFloatRegex),
                val='DoubleFr'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    FrenchNumeric.DoubleExponentialNotationRegex),
                val='DoublePow'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    FrenchNumeric.DoubleCaretExponentialNotationRegex),
                val='DoublePow')
        ]


class FrenchFractionExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[NamedTuple('re_val', [('re', Pattern), ('val', str)])]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_FRACTION

    def __init__(self, mode):
        self.__regexes = [
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    FrenchNumeric.FractionNotationWithSpacesRegex),
                val='FracNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    FrenchNumeric.FractionNotationRegex),
                val='FracNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    FrenchNumeric.FractionNounRegex),
                val='FracFr'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    FrenchNumeric.FractionNounWithArticleRegex),
                val='FracFr')
        ]

        if mode != NumberMode.Unit:
            self.__regexes.append(
                ReVal(
                    re=RegExpUtility.get_safe_reg_exp(
                        FrenchNumeric.FractionPrepositionRegex),
                    val='FracFr'))


class FrenchOrdinalExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[NamedTuple('re_val', [('re', Pattern), ('val', str)])]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_ORDINAL

    def __init__(self):
        self.__regexes = [
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    FrenchNumeric.OrdinalSuffixRegex),
                val='OrdinalNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    FrenchNumeric.OrdinalFrenchRegex),
                val='OrdFr')
        ]


class FrenchPercentageExtractor(BasePercentageExtractor):
    def __init__(self):
        super().__init__(FrenchNumberExtractor(NumberMode.DEFAULT))

    def get_definitions(self) -> List[str]:
        return [
            FrenchNumeric.NumberWithSuffixPercentage,
            FrenchNumeric.NumberWithPrefixPercentage
        ]



from typing import Pattern, List, NamedTuple
import regex

from recognizers_text.utilities import RegExpUtility
from recognizers_number.number.models import NumberMode, LongFormatMode
from recognizers_number.resources import BaseNumbers
from recognizers_number.resources.english_numeric import EnglishNumeric
from recognizers_number.number.extractors import ReVal, ReRe, BaseNumberExtractor, BasePercentageExtractor
from recognizers_number.number.constants import Constants


class EnglishNumberExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[ReVal]:
        return self.__regexes

    @property
    def ambiguity_filters_dict(self) -> List[ReRe]:
        return self.__ambiguity_filters_dict

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM

    @property
    def _negative_number_terms(self) -> Pattern:
        return self.__negative_number_terms

    def __init__(self, mode: NumberMode = NumberMode.DEFAULT):
        self.__negative_number_terms = RegExpUtility.get_safe_reg_exp(
            EnglishNumeric.NegativeNumberTermsRegex)
        self.__regexes: List[ReVal] = list()
        cardinal_ex: EnglishCardinalExtractor = None

        if mode is NumberMode.PURE_NUMBER:
            cardinal_ex = EnglishCardinalExtractor(
                EnglishNumeric.PlaceHolderPureNumber)
        elif mode is NumberMode.CURRENCY:
            self.__regexes.append(ReVal(re=RegExpUtility.get_safe_reg_exp(
                EnglishNumeric.CurrencyRegex), val='IntegerNum'))

        if cardinal_ex is None:
            cardinal_ex = EnglishCardinalExtractor()

        self.__regexes.extend(cardinal_ex.regexes)

        fraction_ex = EnglishFractionExtractor(mode)
        self.__regexes.extend(fraction_ex.regexes)

        ambiguity_filters_dict: List[ReRe] = list()

        if mode != NumberMode.Unit:
            for key, value in EnglishNumeric.AmbiguityFiltersDict.items():
                ambiguity_filters_dict.append(ReRe(reKey=RegExpUtility.get_safe_reg_exp(key),
                                                   reVal=RegExpUtility.get_safe_reg_exp(value)))
        self.__ambiguity_filters_dict = ambiguity_filters_dict


class EnglishCardinalExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[ReVal]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_CARDINAL

    def __init__(self, placeholder: str = EnglishNumeric.PlaceHolderDefault):
        self.__regexes: List[ReVal] = list()

        # Add integer regexes
        integer_ex = EnglishIntegerExtractor(placeholder)
        self.__regexes.extend(integer_ex.regexes)

        # Add double regexes
        double_ex = EnglishDoubleExtractor(placeholder)
        self.__regexes.extend(double_ex.regexes)


class EnglishIntegerExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[NamedTuple('re_val', [('re', Pattern), ('val', str)])]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_INTEGER

    def __init__(self, placeholder: str = EnglishNumeric.PlaceHolderDefault):
        self.__regexes = [
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    EnglishNumeric.NumbersWithPlaceHolder(placeholder)),
                val='IntegerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    EnglishNumeric.NumbersWithSuffix, regex.S),
                val='IntegerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(self._generate_format_regex(
                    LongFormatMode.INTEGER_COMMA, placeholder)),
                val='IntegerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(self._generate_format_regex(
                    LongFormatMode.INTEGER_BLANK, placeholder)),
                val='IntegerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(self._generate_format_regex(
                    LongFormatMode.INTEGER_NO_BREAK_SPACE, placeholder)),
                val='IntegerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    EnglishNumeric.RoundNumberIntegerRegexWithLocks),
                val='IntegerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    EnglishNumeric.NumbersWithDozenSuffix),
                val='IntegerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    EnglishNumeric.AllIntRegexWithLocks),
                val='IntegerEng'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    EnglishNumeric.AllIntRegexWithDozenSuffixLocks),
                val='IntegerEng')
        ]


class EnglishDoubleExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[NamedTuple('re_val', [('re', Pattern), ('val', str)])]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_DOUBLE

    def __init__(self, placeholder):
        self.__regexes = [
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    EnglishNumeric.DoubleDecimalPointRegex(placeholder)),
                val='DoubleNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    EnglishNumeric.DoubleWithoutIntegralRegex(placeholder)),
                val='DoubleNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(self._generate_format_regex(
                    LongFormatMode.DOUBLE_COMMA_DOT, placeholder)),
                val='DoubleNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(self._generate_format_regex(
                    LongFormatMode.DOUBLE_NO_BREAK_SPACE_DOT, placeholder)),
                val='DoubleNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    EnglishNumeric.DoubleWithMultiplierRegex, regex.S),
                val='DoubleNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    EnglishNumeric.DoubleWithRoundNumber),
                val='DoubleNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    EnglishNumeric.DoubleAllFloatRegex),
                val='DoubleEng'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    EnglishNumeric.DoubleExponentialNotationRegex),
                val='DoublePow'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    EnglishNumeric.DoubleCaretExponentialNotationRegex),
                val='DoublePow')
        ]


class EnglishFractionExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[NamedTuple('re_val', [('re', Pattern), ('val', str)])]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_FRACTION

    def __init__(self, mode):
        self.__regexes = [
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    EnglishNumeric.FractionNotationWithSpacesRegex),
                val='FracNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    EnglishNumeric.FractionNotationRegex),
                val='FracNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    EnglishNumeric.FractionNounRegex),
                val='FracEng'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    EnglishNumeric.FractionNounWithArticleRegex),
                val='FracEng')
        ]

        if mode != NumberMode.Unit:
            self.__regexes.append(
                ReVal(
                    re=RegExpUtility.get_safe_reg_exp(
                        EnglishNumeric.FractionPrepositionRegex),
                    val='FracEng'))


class EnglishOrdinalExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[NamedTuple('re_val', [('re', Pattern), ('val', str)])]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_ORDINAL

    def __init__(self):
        self.__regexes = [
            ReVal(
                re=EnglishNumeric.OrdinalSuffixRegex,
                val='OrdinalNum'),
            ReVal(
                re=EnglishNumeric.OrdinalNumericRegex,
                val='OrdinalNum'),
            ReVal(
                re=EnglishNumeric.OrdinalEnglishRegex,
                val='OrdEng'),
            ReVal(
                re=EnglishNumeric.OrdinalRoundNumberRegex,
                val='OrdEng')
        ]


class EnglishPercentageExtractor(BasePercentageExtractor):
    def __init__(self):
        super().__init__(EnglishNumberExtractor(NumberMode.DEFAULT))

    def get_definitions(self) -> List[str]:
        return [
            EnglishNumeric.NumberWithSuffixPercentage,
            EnglishNumeric.NumberWithPrefixPercentage
        ]



class Constants:
    SYS_NUM_CARDINAL: str = 'builtin.num.cardinal'
    SYS_NUM_DOUBLE: str = 'builtin.num.double'
    SYS_NUM_FRACTION: str = 'builtin.num.fraction'
    SYS_NUM_INTEGER: str = 'builtin.num.integer'
    SYS_NUM: str = 'builtin.num'
    SYS_NUM_ORDINAL: str = 'builtin.num.ordinal'
    SYS_NUM_PERCENTAGE: str = 'builtin.num.percentage'

    # NARROW NO-BREAK SPACE
    NO_BREAK_SPACE: chr = '\u202f'



from typing import Pattern, List, NamedTuple

from recognizers_text.utilities import RegExpUtility
from recognizers_number.number.models import NumberMode, LongFormatMode
from recognizers_number.resources import BaseNumbers
from recognizers_number.resources.spanish_numeric import SpanishNumeric
from recognizers_number.number.extractors import ReVal, ReRe, BaseNumberExtractor, BasePercentageExtractor
from recognizers_number.number.constants import Constants


class SpanishNumberExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[ReVal]:
        return self.__regexes

    @property
    def ambiguity_filters_dict(self) -> List[ReRe]:
        return self.__ambiguity_filters_dict

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM

    def __init__(self, mode: NumberMode = NumberMode.DEFAULT):
        self.__regexes: List[ReVal] = list()
        cardinal_ex: SpanishCardinalExtractor = None

        if mode is NumberMode.PURE_NUMBER:
            cardinal_ex = SpanishCardinalExtractor(
                SpanishNumeric.PlaceHolderPureNumber)
        elif mode is NumberMode.CURRENCY:
            self.__regexes.append(
                ReVal(re=SpanishNumeric.CurrencyRegex, val='IntegerNum'))

        if cardinal_ex is None:
            cardinal_ex = SpanishCardinalExtractor()

        self.__regexes.extend(cardinal_ex.regexes)

        fraction_ex = SpanishFractionExtractor(mode)
        self.__regexes.extend(fraction_ex.regexes)

        ambiguity_filters_dict: List[ReRe] = list()

        if mode != NumberMode.Unit:
            for key, value in SpanishNumeric.AmbiguityFiltersDict.items():
                ambiguity_filters_dict.append(ReRe(reKey=RegExpUtility.get_safe_reg_exp(key),
                                                   reVal=RegExpUtility.get_safe_reg_exp(value)))
        self.__ambiguity_filters_dict = ambiguity_filters_dict


class SpanishCardinalExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[ReVal]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_CARDINAL

    def __init__(self, placeholder: str = SpanishNumeric.PlaceHolderDefault):
        self.__regexes: List[ReVal] = list()

        # Add integer regexes
        integer_ex = SpanishIntegerExtractor(placeholder)
        self.__regexes.extend(integer_ex.regexes)

        # Add double regexes
        double_ex = SpanishDoubleExtractor(placeholder)
        self.__regexes.extend(double_ex.regexes)


class SpanishIntegerExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[
            NamedTuple('re_val', [('re', Pattern), ('val', str)])]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_INTEGER

    def __init__(self, placeholder: str = SpanishNumeric.PlaceHolderDefault):
        self.__regexes = [
            ReVal(
                re=SpanishNumeric.NumbersWithPlaceHolder(placeholder),
                val='IntegerNum'),
            ReVal(
                re=SpanishNumeric.NumbersWithSuffix,
                val='IntegerNum'),
            ReVal(
                re=self._generate_format_regex(LongFormatMode.INTEGER_DOT,
                                               placeholder),
                val='IntegerNum'),
            ReVal(
                re=self._generate_format_regex(LongFormatMode.INTEGER_BLANK,
                                               placeholder),
                val='IntegerNum'),
            ReVal(
                re=self._generate_format_regex(
                    LongFormatMode.INTEGER_NO_BREAK_SPACE, placeholder),
                val='IntegerNum'),
            ReVal(
                re=SpanishNumeric.RoundNumberIntegerRegexWithLocks,
                val='IntegerNum'),
            ReVal(
                re=SpanishNumeric.NumbersWithDozenSuffix,
                val='IntegerNum'),
            ReVal(
                re=SpanishNumeric.AllIntRegexWithLocks,
                val='IntegerSpa'),
            ReVal(
                re=SpanishNumeric.AllIntRegexWithDozenSuffixLocks,
                val='IntegerSpa')
        ]


class SpanishDoubleExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[
            NamedTuple('re_val', [('re', Pattern), ('val', str)])]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_DOUBLE

    def __init__(self, placeholder: str = SpanishNumeric.PlaceHolderDefault):
        self.__regexes = [
            ReVal(
                re=SpanishNumeric.DoubleDecimalPointRegex(placeholder),
                val='DoubleNum'),
            ReVal(
                re=SpanishNumeric.DoubleWithoutIntegralRegex(placeholder),
                val='DoubleNum'),
            ReVal(
                re=SpanishNumeric.DoubleWithMultiplierRegex,
                val='DoubleNum'),
            ReVal(
                re=SpanishNumeric.DoubleWithRoundNumber,
                val='DoubleNum'),
            ReVal(
                re=SpanishNumeric.DoubleAllFloatRegex,
                val='DoubleSpa'),
            ReVal(
                re=SpanishNumeric.DoubleExponentialNotationRegex,
                val='DoublePow'),
            ReVal(
                re=SpanishNumeric.DoubleCaretExponentialNotationRegex,
                val='DoublePow'),
            ReVal(
                re=self._generate_format_regex(LongFormatMode.DOUBLE_DOT_COMMA,
                                               placeholder),
                val='DoubleNum'),
            ReVal(
                re=self._generate_format_regex(
                    LongFormatMode.DOUBLE_NO_BREAK_SPACE_COMMA,
                    placeholder),
                val='DoubleNum')
        ]


class SpanishFractionExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[
            NamedTuple('re_val', [('re', Pattern), ('val', str)])]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_FRACTION

    def __init__(self, mode):
        self.__regexes = [
            ReVal(
                re=SpanishNumeric.FractionNotationRegex,
                val='FracNum'),
            ReVal(
                re=SpanishNumeric.FractionNotationWithSpacesRegex,
                val='FracNum'),
            ReVal(
                re=SpanishNumeric.FractionNounRegex,
                val='FracSpa'),
            ReVal(
                re=SpanishNumeric.FractionNounWithArticleRegex,
                val='FracSpa')
        ]

        if mode != NumberMode.Unit:
            self.__regexes.append(
                ReVal(
                    re=SpanishNumeric.FractionPrepositionRegex,
                    val='FracSpa'))


class SpanishOrdinalExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[
            NamedTuple('re_val', [('re', Pattern), ('val', str)])]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_ORDINAL

    def __init__(self):
        self.__regexes = [
            ReVal(
                re=SpanishNumeric.OrdinalSuffixRegex,
                val='OrdinalNum'),
            ReVal(
                re=SpanishNumeric.OrdinalNounRegex,
                val='OrdSpa')
        ]


class SpanishPercentageExtractor(BasePercentageExtractor):
    def __init__(self):
        super().__init__(SpanishNumberExtractor())

    def get_definitions(self) -> List[str]:
        return [
            SpanishNumeric.NumberWithPrefixPercentage
        ]



# ------------------------------------------------------------------------------
# <auto-generated>
#     This code was generated by a tool.
#     Changes to this file may cause incorrect behavior and will be lost if
#     the code is regenerated.
# </auto-generated>
#
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# ------------------------------------------------------------------------------

# pylint: disable=line-too-long


class BaseNumbers:
    NumberReplaceToken = '@builtin.num'
    FractionNumberReplaceToken = '@builtin.num.fraction'

    def IntegerRegexDefinition(placeholder, thousandsmark):
        return f'(((?<!\\d+\\s*)-\\s*)|((?<=\\b)(?<!(\\d+\\.|\\d+,))))\\d{{1,3}}({thousandsmark}\\d{{3}})+(?={placeholder})'

    def DoubleRegexDefinition(placeholder, thousandsmark, decimalmark):
        return f'(((?<!\\d+\\s*)-\\s*)|((?<=\\b)(?<!\\d+\\.|\\d+,)))\\d{{1,3}}({thousandsmark}\\d{{3}})+{decimalmark}\\d+(?={placeholder})'
    PlaceHolderDefault = '\\D|\\b'
    NumberMultiplierRegex = f'(K|k|M|G|T|B|b)'
    MultiplierLookupRegex = f'(k|m|t|g|b)'
    CurrencyRegex = f'(((?<=\\W|^)-\\s*)|(?<=\\b))\\d+\\s*(b|m|t|g)(?=\\b)'
    CommonCurrencySymbol = f'(¥|\\$|€|£|₩)'
# pylint: enable=line-too-long



from abc import abstractmethod
from typing import List, Optional
from enum import Enum
from collections import namedtuple

from recognizers_text import Model, ModelResult
from recognizers_text.extractor import Extractor, ExtractResult
from recognizers_text.parser import Parser
from recognizers_text.utilities import QueryProcessor
from recognizers_number.number.constants import Constants


class NumberMode(Enum):
    DEFAULT = 0
    CURRENCY = 1
    PURE_NUMBER = 2
    Unit = 3


LongFormatType = namedtuple(
    'LongFormatType', ['thousands_mark', 'decimals_mark'])


class LongFormatMode:
    INTEGER_COMMA = LongFormatType(thousands_mark=',', decimals_mark=None)
    INTEGER_DOT = LongFormatType(thousands_mark='.', decimals_mark=None)
    INTEGER_BLANK = LongFormatType(thousands_mark=' ', decimals_mark=None)
    INTEGER_NO_BREAK_SPACE = LongFormatType(
        thousands_mark=Constants.NO_BREAK_SPACE, decimals_mark=None)
    DOUBLE_COMMA_DOT = LongFormatType(thousands_mark=',', decimals_mark='.')
    DOUBLE_NO_BREAK_SPACE_DOT = LongFormatType(
        thousands_mark=Constants.NO_BREAK_SPACE, decimals_mark='.')
    DOUBLE_DOT_COMMA = LongFormatType(thousands_mark='.', decimals_mark=',')
    DOUBLE_NO_BREAK_SPACE_COMMA = LongFormatType(
        thousands_mark=Constants.NO_BREAK_SPACE, decimals_mark=',')


class AbstractNumberModel(Model):
    @property
    @abstractmethod
    def model_type_name(self) -> str:
        raise NotImplementedError

    def __init__(self, parser: Parser, extractor: Extractor):
        self.parser: Parser = parser
        self.extractor: Extractor = extractor

    def parse(self, query: str) -> List[ModelResult]:
        query = QueryProcessor.preprocess(query, True)
        results = []

        try:
            extract_results = self.extractor.extract(query)
            results = list(map(self.__single_parse, extract_results))
        except Exception:
            pass

        return results

    def __single_parse(self, source: ExtractResult) -> Optional[ModelResult]:
        parse_result = self.parser.parse(source)
        if parse_result is None:
            return None
        result = ModelResult()
        result.start = parse_result.start
        result.end = parse_result.end
        result.resolution = dict([('value', parse_result.resolution_str)])
        result.text = parse_result.text
        result.type_name = self.model_type_name
        return result


class NumberModel(AbstractNumberModel):
    @property
    def model_type_name(self) -> str:
        return 'number'


class OrdinalModel(AbstractNumberModel):
    @property
    def model_type_name(self) -> str:
        return 'ordinal'


class PercentModel(AbstractNumberModel):
    @property
    def model_type_name(self) -> str:
        return 'percentage'



# ------------------------------------------------------------------------------
# <auto-generated>
#     This code was generated by a tool.
#     Changes to this file may cause incorrect behavior and will be lost if
#     the code is regenerated.
# </auto-generated>
#
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# ------------------------------------------------------------------------------

from .base_numbers import BaseNumbers
# pylint: disable=line-too-long


class FrenchNumeric:
    LangMarker = 'Fr'
    CompoundNumberLanguage = True
    MultiDecimalSeparatorCulture = True
    RoundNumberIntegerRegex = f'(cent|mille|millions|million|milliard|milliards|billion|billions)'
    ZeroToNineIntegerRegex = f'(et un|un|une|deux|trois|quatre|cinq|six|sept|huit|neuf)'
    TenToNineteenIntegerRegex = f'((seize|quinze|quatorze|treize|douze|onze)|dix(\\Wneuf|\\Whuit|\\Wsept)?)'
    TensNumberIntegerRegex = f'(quatre\\Wvingt(s|\\Wdix)?|soixante\\Wdix|vingt|trente|quarante|cinquante|soixante|septante|octante|huitante|nonante)'
    DigitsNumberRegex = f'\\d|\\d{{1,3}}(\\.\\d{{3}})'
    NegativeNumberTermsRegex = f'^[.]'
    NegativeNumberSignRegex = f'^({NegativeNumberTermsRegex}\\s+).*'
    HundredsNumberIntegerRegex = f'(({ZeroToNineIntegerRegex}(\\s+cent))|cent|((\\s+cent\\s)+{TensNumberIntegerRegex}))'
    BelowHundredsRegex = f'(({TenToNineteenIntegerRegex}|({TensNumberIntegerRegex}([-\\s]+({TenToNineteenIntegerRegex}|{ZeroToNineIntegerRegex}))?))|{ZeroToNineIntegerRegex})'
    BelowThousandsRegex = f'(({HundredsNumberIntegerRegex}(\\s+{BelowHundredsRegex})?|{BelowHundredsRegex}|{TenToNineteenIntegerRegex})|cent\\s+{TenToNineteenIntegerRegex})'
    SupportThousandsRegex = f'(({BelowThousandsRegex}|{BelowHundredsRegex})\\s+{RoundNumberIntegerRegex}(\\s+{RoundNumberIntegerRegex})?)'
    SeparaIntRegex = f'({SupportThousandsRegex}(\\s+{SupportThousandsRegex})*(\\s+{BelowThousandsRegex})?|{BelowThousandsRegex})'
    AllIntRegex = f'({SeparaIntRegex}|mille(\\s+{BelowThousandsRegex})?)'

    def NumbersWithPlaceHolder(placeholder):
        return f'(((?<!\\d+\\s*)-\\s*)|(?<=\\b))\\d+(?!([,\\.]\\d+[a-zA-Z]))(?={placeholder})'
    NumbersWithSuffix = f'(((?<=\\W|^)-\\s*)|(?<=\\b))\\d+\\s*{BaseNumbers.NumberMultiplierRegex}(?=\\b)'
    RoundNumberIntegerRegexWithLocks = f'(?<=\\b)({DigitsNumberRegex})+\\s+{RoundNumberIntegerRegex}(?=\\b)'
    NumbersWithDozenSuffix = f'(((?<!\\d+\\s*)-\\s*)|(?<=\\b))\\d+\\s+douzaine(s)?(?=\\b)'
    AllIntRegexWithLocks = f'((?<=\\b){AllIntRegex}(?=\\b))'
    AllIntRegexWithDozenSuffixLocks = f'(?<=\\b)(((demi\\s+)?\\s+douzaine)|({AllIntRegex}\\s+douzaines?))(?=\\b)'
    SimpleRoundOrdinalRegex = f'(centi[eè]me|milli[eè]me|millioni[eè]me|milliardi[eè]me|billioni[eè]me)'
    OneToNineOrdinalRegex = f'(premier|premi[eè]re|deuxi[eè]me|second[e]|troisi[eè]me|tiers|tierce|quatri[eè]me|cinqui[eè]me|sixi[eè]me|septi[eè]me|huiti[eè]me|neuvi[eè]me)'
    SpecialUnderHundredOrdinalRegex = f'(onzi[eè]me|douzi[eè]me)'
    TensOrdinalRegex = f'(quatre-vingt-dixi[eè]me|quatre-vingti[eè]me|huitanti[eè]me|octanti[eè]me|soixante-dixi[eè]me|septanti[eè]me|soixanti[eè]me|cinquanti[eè]me|quaranti[eè]me|trenti[eè]me|vingti[eè]me)'
    HundredOrdinalRegex = f'({AllIntRegex}(\\s+(centi[eè]me\\s)))'
    UnderHundredOrdinalRegex = f'((({AllIntRegex}(\\W)?)?{OneToNineOrdinalRegex})|({TensNumberIntegerRegex}(\\W)?)?{OneToNineOrdinalRegex}|{TensOrdinalRegex}|{SpecialUnderHundredOrdinalRegex})'
    UnderThousandOrdinalRegex = f'((({HundredOrdinalRegex}(\\s)?)?{UnderHundredOrdinalRegex})|(({AllIntRegex}(\\W)?)?{SimpleRoundOrdinalRegex})|{HundredOrdinalRegex})'
    OverThousandOrdinalRegex = f'(({AllIntRegex})(i[eè]me))'
    ComplexOrdinalRegex = f'(({OverThousandOrdinalRegex}(\\s)?)?{UnderThousandOrdinalRegex}|{OverThousandOrdinalRegex}|{UnderHundredOrdinalRegex})'
    SuffixOrdinalRegex = f'(({AllIntRegex})({SimpleRoundOrdinalRegex}))'
    ComplexRoundOrdinalRegex = f'((({SuffixOrdinalRegex}(\\s)?)?{ComplexOrdinalRegex})|{SuffixOrdinalRegex})'
    AllOrdinalRegex = f'({ComplexOrdinalRegex}|{SimpleRoundOrdinalRegex}|{ComplexRoundOrdinalRegex})'
    PlaceHolderPureNumber = f'\\b'
    PlaceHolderDefault = f'\\D|\\b'
    OrdinalSuffixRegex = f'(?<=\\b)((\\d*(1er|2e|2eme|3e|3eme|4e|4eme|5e|5eme|6e|6eme|7e|7eme|8e|8eme|9e|9eme|0e|0eme))|(11e|11eme|12e|12eme))(?=\\b)'
    OrdinalFrenchRegex = f'(?<=\\b){AllOrdinalRegex}(?=\\b)'
    FractionNotationWithSpacesRegex = f'(((?<=\\W|^)-\\s*)|(?<=\\b))\\d+\\s+\\d+[/]\\d+(?=(\\b[^/]|$))'
    FractionNotationRegex = f'(((?<=\\W|^)-\\s*)|(?<=\\b))\\d+[/]\\d+(?=(\\b[^/]|$))'
    FractionNounRegex = f'(?<=\\b)({AllIntRegex}\\s+((et)\\s+)?)?({AllIntRegex})(\\s+((et)\\s)?)((({AllOrdinalRegex})s?|({SuffixOrdinalRegex})s?)|demis?|tiers?|quarts?)(?=\\b)'
    FractionNounWithArticleRegex = f'(?<=\\b)({AllIntRegex}\\s+(et\\s+)?)?(un|une)(\\s+)(({AllOrdinalRegex})|({SuffixOrdinalRegex})|(et\\s+)?demis?)(?=\\b)'
    FractionPrepositionRegex = f'(?<!{BaseNumbers.CommonCurrencySymbol}\\s*)(?<=\\b)(?<numerator>({AllIntRegex})|((?<!\\.)\\d+))\\s+sur\\s+(?<denominator>({AllIntRegex})|((\\d+)(?!\\.)))(?=\\b)'
    AllPointRegex = f'((\\s+{ZeroToNineIntegerRegex})+|(\\s+{SeparaIntRegex}))'
    AllFloatRegex = f'({AllIntRegex}(\\s+(virgule|point)){AllPointRegex})'

    def DoubleDecimalPointRegex(placeholder):
        return f'(((?<!\\d+\\s*)-\\s*)|((?<=\\b)(?<!\\d+[,\\.])))\\d+[,\\.]\\d+(?!([,\\.]\\d+))(?={placeholder})'

    def DoubleWithoutIntegralRegex(placeholder):
        return f'(?<=\\s|^)(?<!(\\d+))[,\\.]\\d+(?!([,\\.]\\d+))(?={placeholder})'
    DoubleWithMultiplierRegex = f'(((?<!\\d+\\s*)-\\s*)|((?<=\\b)(?<!\\d+\\[,\\.])))\\d+[,\\.]\\d+\\s*{BaseNumbers.NumberMultiplierRegex}(?=\\b)'
    DoubleWithRoundNumber = f'(((?<!\\d+\\s*)-\\s*)|((?<=\\b)(?<!\\d+\\[,\\.])))\\d+[,\\.]\\d+\\s+{RoundNumberIntegerRegex}(?=\\b)'
    DoubleAllFloatRegex = f'((?<=\\b){AllFloatRegex}(?=\\b))'
    DoubleExponentialNotationRegex = f'(((?<!\\d+\\s*)-\\s*)|((?<=\\b)(?<!\\d+[,\\.])))(\\d+([,\\.]\\d+)?)e([+-]*[1-9]\\d*)(?=\\b)'
    DoubleCaretExponentialNotationRegex = f'(((?<!\\d+\\s*)-\\s*)|((?<=\\b)(?<!\\d+[,\\.])))(\\d+([,\\.]\\d+)?)\\^([+-]*[1-9]\\d*)(?=\\b)'
    NumberWithSuffixPercentage = f'(?<!%)({BaseNumbers.NumberReplaceToken})(\\s*)(%(?!{BaseNumbers.NumberReplaceToken})|(pourcentages|pourcents|pourcentage|pourcent)\\b)'
    NumberWithPrefixPercentage = f'((?<!{BaseNumbers.NumberReplaceToken})%|pourcent|pourcent des|pourcentage de)(\\s*)({BaseNumbers.NumberReplaceToken})(?=\\s|$)'
    DecimalSeparatorChar = ','
    FractionMarkerToken = 'sur'
    NonDecimalSeparatorChar = '.'
    HalfADozenText = 'six'
    WordSeparatorToken = 'et'
    WrittenDecimalSeparatorTexts = [r'virgule']
    WrittenGroupSeparatorTexts = [r'point', r'points']
    WrittenIntegerSeparatorTexts = [r'et', r'-']
    WrittenFractionSeparatorTexts = [r'et', r'sur']
    HalfADozenRegex = f'(?<=\\b)demi\\s+douzaine'
    DigitalNumberRegex = f'((?<=\\b)(cent|mille|million|millions|milliard|milliards|billions|billion|douzaine(s)?)(?=\\b))|((?<=(\\d|\\b)){BaseNumbers.MultiplierLookupRegex}(?=\\b))'
    AmbiguousFractionConnectorsRegex = f'^[.]'
    CardinalNumberMap = dict([("zéro", 0),
                              ("zero", 0),
                              ("un", 1),
                              ("une", 1),
                              ("deux", 2),
                              ("trois", 3),
                              ("quatre", 4),
                              ("cinq", 5),
                              ("six", 6),
                              ("sept", 7),
                              ("huit", 8),
                              ("neuf", 9),
                              ("dix", 10),
                              ("onze", 11),
                              ("douze", 12),
                              ("treize", 13),
                              ("quatorze", 14),
                              ("quinze", 15),
                              ("seize", 16),
                              ("dix-sept", 17),
                              ("dix-huit", 18),
                              ("dix-neuf", 19),
                              ("vingt", 20),
                              ("trente", 30),
                              ("quarante", 40),
                              ("cinquante", 50),
                              ("soixante", 60),
                              ("soixante-dix", 70),
                              ("septante", 70),
                              ("quatre-vingts", 80),
                              ("quatre-vingt", 80),
                              ("quatre vingts", 80),
                              ("quatre vingt", 80),
                              ("quatre-vingt-dix", 90),
                              ("quatre-vingt dix", 90),
                              ("quatre vingt dix", 90),
                              ("quatre-vingts-dix", 90),
                              ("quatre-vingts-onze", 91),
                              ("quatre-vingt-onze", 91),
                              ("quatre-vingts-douze", 92),
                              ("quatre-vingt-douze", 92),
                              ("quatre-vingts-treize", 93),
                              ("quatre-vingt-treize", 93),
                              ("quatre-vingts-quatorze", 94),
                              ("quatre-vingt-quatorze", 94),
                              ("quatre-vingts-quinze", 95),
                              ("quatre-vingt-quinze", 95),
                              ("quatre-vingts-seize", 96),
                              ("quatre-vingt-seize", 96),
                              ("huitante", 80),
                              ("octante", 80),
                              ("nonante", 90),
                              ("cent", 100),
                              ("mille", 1000),
                              ("un million", 1000000),
                              ("million", 1000000),
                              ("millions", 1000000),
                              ("un milliard", 1000000000),
                              ("milliard", 1000000000),
                              ("milliards", 1000000000),
                              ("un mille milliards", 1000000000000),
                              ("un billion", 1000000000000)])
    OrdinalNumberMap = dict([("premier", 1),
                             ("première", 1),
                             ("premiere", 1),
                             ("deuxième", 2),
                             ("deuxieme", 2),
                             ("second", 2),
                             ("seconde", 2),
                             ("troisième", 3),
                             ("demi", 2),
                             ("tiers", 3),
                             ("tierce", 3),
                             ("quart", 4),
                             ("quarts", 4),
                             ("troisieme", 3),
                             ("quatrième", 4),
                             ("quatrieme", 4),
                             ("cinquième", 5),
                             ("cinquieme", 5),
                             ("sixième", 6),
                             ("sixieme", 6),
                             ("septième", 7),
                             ("septieme", 7),
                             ("huitième", 8),
                             ("huitieme", 8),
                             ("neuvième", 9),
                             ("neuvieme", 9),
                             ("dixième", 10),
                             ("dixieme", 10),
                             ("onzième", 11),
                             ("onzieme", 11),
                             ("douzième", 12),
                             ("douzieme", 12),
                             ("treizième", 13),
                             ("treizieme", 13),
                             ("quatorzième", 14),
                             ("quatorizieme", 14),
                             ("quinzième", 15),
                             ("quinzieme", 15),
                             ("seizième", 16),
                             ("seizieme", 16),
                             ("dix-septième", 17),
                             ("dix-septieme", 17),
                             ("dix-huitième", 18),
                             ("dix-huitieme", 18),
                             ("dix-neuvième", 19),
                             ("dix-neuvieme", 19),
                             ("vingtième", 20),
                             ("vingtieme", 20),
                             ("trentième", 30),
                             ("trentieme", 30),
                             ("quarantième", 40),
                             ("quarantieme", 40),
                             ("cinquantième", 50),
                             ("cinquantieme", 50),
                             ("soixantième", 60),
                             ("soixantieme", 60),
                             ("soixante-dixième", 70),
                             ("soixante-dixieme", 70),
                             ("septantième", 70),
                             ("septantieme", 70),
                             ("quatre-vingtième", 80),
                             ("quatre-vingtieme", 80),
                             ("huitantième", 80),
                             ("huitantieme", 80),
                             ("octantième", 80),
                             ("octantieme", 80),
                             ("quatre-vingt-dixième", 90),
                             ("quatre-vingt-dixieme", 90),
                             ("nonantième", 90),
                             ("nonantieme", 90),
                             ("centième", 100),
                             ("centieme", 100),
                             ("millième", 1000),
                             ("millieme", 1000),
                             ("millionième", 1000000),
                             ("millionieme", 1000000),
                             ("milliardième", 1000000000),
                             ("milliardieme", 1000000000),
                             ("billionieme", 1000000000000),
                             ("billionième", 1000000000000),
                             ("trillionième", 1000000000000000000),
                             ("trillionieme", 1000000000000000000)])
    PrefixCardinalMap = dict([("deux", 2),
                              ("trois", 3),
                              ("quatre", 4),
                              ("cinq", 5),
                              ("six", 6),
                              ("sept", 7),
                              ("huit", 8),
                              ("neuf", 9),
                              ("dix", 10),
                              ("onze", 11),
                              ("douze", 12),
                              ("treize", 13),
                              ("quatorze", 14),
                              ("quinze", 15),
                              ("seize", 16),
                              ("dix sept", 17),
                              ("dix-sept", 17),
                              ("dix-huit", 18),
                              ("dix huit", 18),
                              ("dix-neuf", 19),
                              ("dix neuf", 19),
                              ("vingt", 20),
                              ("vingt-et-un", 21),
                              ("vingt et un", 21),
                              ("vingt-deux", 21),
                              ("vingt deux", 22),
                              ("vingt-trois", 23),
                              ("vingt trois", 23),
                              ("vingt-quatre", 24),
                              ("vingt quatre", 24),
                              ("vingt-cinq", 25),
                              ("vingt cinq", 25),
                              ("vingt-six", 26),
                              ("vingt six", 26),
                              ("vingt-sept", 27),
                              ("vingt sept", 27),
                              ("vingt-huit", 28),
                              ("vingt huit", 28),
                              ("vingt-neuf", 29),
                              ("vingt neuf", 29),
                              ("trente", 30),
                              ("quarante", 40),
                              ("cinquante", 50),
                              ("soixante", 60),
                              ("soixante-dix", 70),
                              ("soixante dix", 70),
                              ("septante", 70),
                              ("quatre-vingt", 80),
                              ("quatre vingt", 80),
                              ("huitante", 80),
                              ("octante", 80),
                              ("nonante", 90),
                              ("quatre vingt dix", 90),
                              ("quatre-vingt-dix", 90),
                              ("cent", 100),
                              ("deux cent", 200),
                              ("trois cents", 300),
                              ("quatre cents", 400),
                              ("cinq cent", 500),
                              ("six cent", 600),
                              ("sept cent", 700),
                              ("huit cent", 800),
                              ("neuf cent", 900)])
    SuffixOrdinalMap = dict([("millième", 1000),
                             ("million", 1000000),
                             ("milliardième", 1000000000000)])
    RoundNumberMap = dict([("cent", 100),
                           ("mille", 1000),
                           ("million", 1000000),
                           ("millions", 1000000),
                           ("milliard", 1000000000),
                           ("milliards", 1000000000),
                           ("billion", 1000000000000),
                           ("billions", 1000000000000),
                           ("centieme", 100),
                           ("centième", 100),
                           ("millieme", 1000),
                           ("millième", 1000),
                           ("millionième", 1000000),
                           ("millionieme", 1000000),
                           ("milliardième", 1000000000),
                           ("milliardieme", 1000000000),
                           ("billionième", 1000000000000),
                           ("billionieme", 1000000000000),
                           ("centiemes", 100),
                           ("centièmes", 100),
                           ("millièmes", 1000),
                           ("milliemes", 1000),
                           ("millionièmes", 1000000),
                           ("millioniemes", 1000000),
                           ("milliardièmes", 1000000000),
                           ("milliardiemes", 1000000000),
                           ("billionièmes", 1000000000000),
                           ("billioniemes", 1000000000000),
                           ("douzaine", 12),
                           ("douzaines", 12),
                           ("k", 1000),
                           ("m", 1000000),
                           ("g", 1000000000),
                           ("b", 1000000000),
                           ("t", 1000000000000)])
    AmbiguityFiltersDict = dict([("^[.]", "")])
    RelativeReferenceOffsetMap = dict([("", "")])
    RelativeReferenceRelativeToMap = dict([("", "")])
# pylint: enable=line-too-long



from recognizers_text.culture import BaseCultureInfo, Culture
from .number import LongFormatMode, LongFormatType

SUPPORTED_CULTURES = {
    Culture.English: LongFormatMode.DOUBLE_COMMA_DOT,
    Culture.Chinese: None,
    Culture.Spanish: LongFormatMode.DOUBLE_DOT_COMMA,
    Culture.Portuguese: LongFormatMode.DOUBLE_DOT_COMMA,
    Culture.French: LongFormatMode.DOUBLE_DOT_COMMA,
    Culture.Japanese: LongFormatMode.DOUBLE_COMMA_DOT
}


class CultureInfo(BaseCultureInfo):
    def format(self, value: object) -> str:
        result = str(value)
        result = result.replace('e', 'E')
        if '.' in result:
            result = result.rstrip('0').rstrip('.')
        if 'E-' in result:
            parts = result.split('E-')
            parts[1] = parts[1].rjust(2, '0')
            result = 'E-'.join(parts)
        if 'E+' in result:
            parts = result.split('E+')
            parts[0] = parts[0].rstrip('0')
            result = 'E+'.join(parts)
        long_format = SUPPORTED_CULTURES.get(self.code)
        if long_format:
            result = ''.join(
                map(lambda x: self.change_mark(x, long_format), result))
        return result

    def change_mark(self, source: str, long_format: LongFormatType) -> str:
        if source == '.':
            return long_format.decimals_mark
        elif source == ',':
            return long_format.thousands_mark
        return source



# ------------------------------------------------------------------------------
# <auto-generated>
#     This code was generated by a tool.
#     Changes to this file may cause incorrect behavior and will be lost if
#     the code is regenerated.
# </auto-generated>
#
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# ------------------------------------------------------------------------------

from .base_numbers import BaseNumbers
# pylint: disable=line-too-long


class PortugueseNumeric:
    LangMarker = 'Por'
    CompoundNumberLanguage = True
    MultiDecimalSeparatorCulture = True
    HundredsNumberIntegerRegex = f'(quatrocent[ao]s|trezent[ao]s|seiscent[ao]s|setecent[ao]s|oitocent[ao]s|novecent[ao]s|duzent[ao]s|quinhent[ao]s|cem|(?<!por\\s+)(cento))'
    RoundNumberIntegerRegex = f'(mil|milh[ãa]o|milh[õo]es|bilh[ãa]o|bilh[õo]es|trilh[ãa]o|trilh[õo]es|qua[td]rilh[ãa]o|qua[td]rilh[õo]es|quintilh[ãa]o|quintilh[õo]es)'
    ZeroToNineIntegerRegex = f'(quatro|cinco|sete|nove|zero|tr[êe]s|seis|oito|dois|duas|um|uma)'
    TenToNineteenIntegerRegex = f'(dez[ea]sseis|dez[ea]ssete|dez[ea]nove|dezoito|quatorze|catorze|quinze|treze|d[ée]z|onze|doze)'
    TensNumberIntegerRegex = f'(cinquenta|quarenta|trinta|sessenta|setenta|oitenta|noventa|vinte)'
    DigitsNumberRegex = f'\\d|\\d{{1,3}}(\\.\\d{{3}})'
    BelowHundredsRegex = f'(({TenToNineteenIntegerRegex}|({TensNumberIntegerRegex}(\\s+e\\s+{ZeroToNineIntegerRegex})?))|{ZeroToNineIntegerRegex})'
    BelowThousandsRegex = f'({HundredsNumberIntegerRegex}(\\s+e\\s+{BelowHundredsRegex})?|{BelowHundredsRegex})'
    SupportThousandsRegex = f'(({BelowThousandsRegex}|{BelowHundredsRegex})\\s+{RoundNumberIntegerRegex}(\\s+{RoundNumberIntegerRegex})?)'
    NegativeNumberTermsRegex = f'^[.]'
    NegativeNumberSignRegex = f'^({NegativeNumberTermsRegex}\\s+).*'
    SeparaIntRegex = f'({SupportThousandsRegex}(\\s+{SupportThousandsRegex})*(\\s+{BelowThousandsRegex})?|{BelowThousandsRegex})'
    AllIntRegex = f'({SeparaIntRegex}|mil(\\s+{BelowThousandsRegex})?)'
    AllPointRegex = f'((\\s+{ZeroToNineIntegerRegex})+|(\\s+{AllIntRegex}))'
    SpecialFractionInteger = f'((({AllIntRegex})i?({ZeroToNineIntegerRegex})|({AllIntRegex}))\\s+a?v[oa]s?)'
    PlaceHolderDefault = f'\\D|\\b'
    PlaceHolderPureNumber = f'\\b'
    AllIntRegexWithLocks = f'((?<=\\b){AllIntRegex}(?=\\b))'
    AllIntRegexWithDozenSuffixLocks = f'(?<=\\b)(((meia)?\\s+(d[úu]zia))|({AllIntRegex}\\s+(e|com)\\s+)?({AllIntRegex}\\s+(d[úu]zia(s)?|dezena(s)?)))(?=\\b)'

    def NumbersWithPlaceHolder(placeholder):
        return f'(((?<!\\d+\\s*)-\\s*)|(?<=\\b))\\d+(?!(,\\d+[a-zA-Z]))(?={placeholder})'
    NumbersWithSuffix = f'(((?<=\\W|^)-\\s*)|(?<=\\b))\\d+\\s*{BaseNumbers.NumberMultiplierRegex}(?=\\b)'
    RoundNumberIntegerRegexWithLocks = f'(?<=\\b)({DigitsNumberRegex})+\\s+{RoundNumberIntegerRegex}(?=\\b)'
    NumbersWithDozenSuffix = f'(((?<!\\d+\\s*)-\\s*)|(?<=\\b))\\d+\\s+dezena(s)?(?=\\b)'
    NumbersWithDozen2Suffix = f'(((?<=\\W|^)-\\s*)|(?<=\\b))\\d+\\s+d[úu]zia(s)(?=\\b)'
    SimpleRoundOrdinalRegex = f'(mil[eé]sim[oa]|milion[eé]sim[oa]|bilion[eé]sim[oa]|trilion[eé]sim[oa]|quatrilion[eé]sim[oa]|quintilion[eé]sim[oa])'
    OneToNineOrdinalRegex = f'(primeir[oa]|segund[oa]|terceir[oa]|quart[oa]|quint[oa]|sext[oa]|s[eé]tim[oa]|oitav[oa]|non[oa])'
    TensOrdinalRegex = f'(nonag[eé]sim[oa]|octog[eé]sim[oa]|setuag[eé]sim[oa]|septuag[eé]sim[oa]|sexag[eé]sim[oa]|quinquag[eé]sim[oa]|quadrag[eé]sim[oa]|trig[eé]sim[oa]|vig[eé]sim[oa]|d[eé]cim[oa])'
    HundredOrdinalRegex = f'(cent[eé]sim[oa]|ducent[eé]sim[oa]|tricent[eé]sim[oa]|cuadringent[eé]sim[oa]|quingent[eé]sim[oa]|sexcent[eé]sim[oa]|septingent[eé]sim[oa]|octingent[eé]sim[oa]|noningent[eé]sim[oa])'
    SpecialUnderHundredOrdinalRegex = f'(und[eé]cim[oa]|duod[eé]cimo)'
    UnderHundredOrdinalRegex = f'((({TensOrdinalRegex}(\\s)?)?{OneToNineOrdinalRegex})|{TensOrdinalRegex}|{SpecialUnderHundredOrdinalRegex})'
    UnderThousandOrdinalRegex = f'((({HundredOrdinalRegex}(\\s)?)?{UnderHundredOrdinalRegex})|{HundredOrdinalRegex})'
    OverThousandOrdinalRegex = f'(({AllIntRegex})([eé]sim[oa]))'
    ComplexOrdinalRegex = f'(({OverThousandOrdinalRegex}(\\s)?)?{UnderThousandOrdinalRegex}|{OverThousandOrdinalRegex})'
    SuffixRoundOrdinalRegex = f'(({AllIntRegex})({SimpleRoundOrdinalRegex}))'
    ComplexRoundOrdinalRegex = f'((({SuffixRoundOrdinalRegex}(\\s)?)?{ComplexOrdinalRegex})|{SuffixRoundOrdinalRegex})'
    AllOrdinalRegex = f'{ComplexOrdinalRegex}|{SimpleRoundOrdinalRegex}|{ComplexRoundOrdinalRegex}'
    OrdinalSuffixRegex = f'(?<=\\b)(\\d*(1[oaº]|2[oaº]|3[oaº]|4[oaº]|5[oaº]|6[oaº]|7[oaº]|8[oaº]|9[oaº]|0[oaº]|1.º|2.º|3.º|4.º|5.º|6.º|7.º|8.º|9.º))(?=\\b)'
    OrdinalEnglishRegex = f'(?<=\\b){AllOrdinalRegex}(?=\\b)'
    FractionNotationRegex = f'(((?<=\\W|^)-\\s*)|(?<=\\b))\\d+[/]\\d+(?=(\\b[^/]|$))'
    FractionNotationWithSpacesRegex = f'(((?<=\\W|^)-\\s*)|(?<=\\b))\\d+\\s+\\d+[/]\\d+(?=(\\b[^/]|$))'
    FractionNounRegex = f'(?<=\\b)({AllIntRegex}\\s+((e|com)\\s+)?)?({AllIntRegex})(\\s+((e|com)\\s)?)((({AllOrdinalRegex})s?|({SpecialFractionInteger})|({SuffixRoundOrdinalRegex})s?)|mei[oa]?|ter[çc]o?)(?=\\b)'
    FractionNounWithArticleRegex = f'(?<=\\b)({AllIntRegex}\\s+(e\\s+)?)?(um|um[as])(\\s+)(({AllOrdinalRegex})|({SuffixRoundOrdinalRegex})|(e\\s+)?mei[oa]?)(?=\\b)'
    FractionPrepositionRegex = f'(?<!{BaseNumbers.CommonCurrencySymbol}\\s*)(?<=\\b)(?<numerator>({AllIntRegex})|((?<!\\.)\\d+))\\s+sobre\\s+(?<denominator>({AllIntRegex})|((\\d+)(?!\\.)))(?=\\b)'
    AllFloatRegex = f'{AllIntRegex}(\\s+(vírgula|virgula|e|ponto)){AllPointRegex}'
    DoubleWithMultiplierRegex = f'(((?<!\\d+\\s*)-\\s*)|((?<=\\b)(?<!\\d+\\,)))\\d+,\\d+\\s*{BaseNumbers.NumberMultiplierRegex}(?=\\b)'
    DoubleExponentialNotationRegex = f'(((?<!\\d+\\s*)-\\s*)|((?<=\\b)(?<!\\d+,)))(\\d+(,\\d+)?)e([+-]*[1-9]\\d*)(?=\\b)'
    DoubleCaretExponentialNotationRegex = f'(((?<!\\d+\\s*)-\\s*)|((?<=\\b)(?<!\\d+,)))(\\d+(,\\d+)?)\\^([+-]*[1-9]\\d*)(?=\\b)'

    def DoubleDecimalPointRegex(placeholder):
        return f'(((?<!\\d+\\s*)-\\s*)|((?<=\\b)(?<!\\d+,)))\\d+,\\d+(?!(,\\d+))(?={placeholder})'

    def DoubleWithoutIntegralRegex(placeholder):
        return f'(?<=\\s|^)(?<!(\\d+)),\\d+(?!(,\\d+))(?={placeholder})'
    DoubleWithRoundNumber = f'(((?<!\\d+\\s*)-\\s*)|((?<=\\b)(?<!\\d+\\,)))\\d+,\\d+\\s+{RoundNumberIntegerRegex}(?=\\b)'
    DoubleAllFloatRegex = f'((?<=\\b){AllFloatRegex}(?=\\b))'
    NumberWithSuffixPercentage = f'(?<!%)({BaseNumbers.NumberReplaceToken})(\\s*)(%(?!{BaseNumbers.NumberReplaceToken})|(por cento|pontos percentuais)\\b)'
    AmbiguousFractionConnectorsRegex = f'^[.]'
    DecimalSeparatorChar = ','
    FractionMarkerToken = 'sobre'
    NonDecimalSeparatorChar = '.'
    HalfADozenText = 'seis'
    WordSeparatorToken = 'e'
    WrittenDecimalSeparatorTexts = [r'virgula', r'vírgula']
    WrittenGroupSeparatorTexts = [r'ponto']
    WrittenIntegerSeparatorTexts = [r'e']
    WrittenFractionSeparatorTexts = [r'com']
    WrittenFractionSuffix = [r'avo', r'ava']
    PluralSuffix = 's'
    HalfADozenRegex = f'meia\\s+d[uú]zia'
    DigitalNumberRegex = f'((?<=\\b)(mil|cem|milh[oõ]es|milh[aã]o|bilh[oõ]es|bilh[aã]o|trilh[oõ]es|trilh[aã]o|milhares|centena|centenas|dezena|dezenas?)(?=\\b))|((?<=(\\d|\\b)){BaseNumbers.MultiplierLookupRegex}(?=\\b))'
    CardinalNumberMap = dict([("zero", 0),
                              ("hum", 1),
                              ("um", 1),
                              ("uma", 1),
                              ("dois", 2),
                              ("duas", 2),
                              ("meia", 2),
                              ("meio", 2),
                              ("tres", 3),
                              ("três", 3),
                              ("quatro", 4),
                              ("cinco", 5),
                              ("seis", 6),
                              ("sete", 7),
                              ("oito", 8),
                              ("nove", 9),
                              ("dez", 10),
                              ("dezena", 10),
                              ("déz", 10),
                              ("onze", 11),
                              ("doze", 12),
                              ("dúzia", 12),
                              ("duzia", 12),
                              ("dúzias", 12),
                              ("duzias", 12),
                              ("treze", 13),
                              ("catorze", 14),
                              ("quatorze", 14),
                              ("quinze", 15),
                              ("dezesseis", 16),
                              ("dezasseis", 16),
                              ("dezessete", 17),
                              ("dezassete", 17),
                              ("dezoito", 18),
                              ("dezenove", 19),
                              ("dezanove", 19),
                              ("vinte", 20),
                              ("trinta", 30),
                              ("quarenta", 40),
                              ("cinquenta", 50),
                              ("cincoenta", 50),
                              ("sessenta", 60),
                              ("setenta", 70),
                              ("oitenta", 80),
                              ("noventa", 90),
                              ("cem", 100),
                              ("cento", 100),
                              ("duzentos", 200),
                              ("duzentas", 200),
                              ("trezentos", 300),
                              ("trezentas", 300),
                              ("quatrocentos", 400),
                              ("quatrocentas", 400),
                              ("quinhentos", 500),
                              ("quinhentas", 500),
                              ("seiscentos", 600),
                              ("seiscentas", 600),
                              ("setecentos", 700),
                              ("setecentas", 700),
                              ("oitocentos", 800),
                              ("oitocentas", 800),
                              ("novecentos", 900),
                              ("novecentas", 900),
                              ("mil", 1000),
                              ("milhão", 1000000),
                              ("milhao", 1000000),
                              ("milhões", 1000000),
                              ("milhoes", 1000000),
                              ("bilhão", 1000000000),
                              ("bilhao", 1000000000),
                              ("bilhões", 1000000000),
                              ("bilhoes", 1000000000),
                              ("trilhão", 1000000000000),
                              ("trilhao", 1000000000000),
                              ("trilhões", 1000000000000),
                              ("trilhoes", 1000000000000)])
    OrdinalNumberMap = dict([("primeiro", 1),
                             ("primeira", 1),
                             ("segundo", 2),
                             ("segunda", 2),
                             ("terceiro", 3),
                             ("terceira", 3),
                             ("quarto", 4),
                             ("quarta", 4),
                             ("quinto", 5),
                             ("quinta", 5),
                             ("sexto", 6),
                             ("sexta", 6),
                             ("sétimo", 7),
                             ("setimo", 7),
                             ("sétima", 7),
                             ("setima", 7),
                             ("oitavo", 8),
                             ("oitava", 8),
                             ("nono", 9),
                             ("nona", 9),
                             ("décimo", 10),
                             ("decimo", 10),
                             ("décima", 10),
                             ("decima", 10),
                             ("undécimo", 11),
                             ("undecimo", 11),
                             ("undécima", 11),
                             ("undecima", 11),
                             ("duodécimo", 11),
                             ("duodecimo", 11),
                             ("duodécima", 11),
                             ("duodecima", 11),
                             ("vigésimo", 20),
                             ("vigesimo", 20),
                             ("vigésima", 20),
                             ("vigesima", 20),
                             ("trigésimo", 30),
                             ("trigesimo", 30),
                             ("trigésima", 30),
                             ("trigesima", 30),
                             ("quadragésimo", 40),
                             ("quadragesimo", 40),
                             ("quadragésima", 40),
                             ("quadragesima", 40),
                             ("quinquagésimo", 50),
                             ("quinquagesimo", 50),
                             ("quinquagésima", 50),
                             ("quinquagesima", 50),
                             ("sexagésimo", 60),
                             ("sexagesimo", 60),
                             ("sexagésima", 60),
                             ("sexagesima", 60),
                             ("septuagésimo", 70),
                             ("septuagesimo", 70),
                             ("septuagésima", 70),
                             ("septuagesima", 70),
                             ("setuagésimo", 70),
                             ("setuagesimo", 70),
                             ("setuagésima", 70),
                             ("setuagesima", 70),
                             ("octogésimo", 80),
                             ("octogesimo", 80),
                             ("octogésima", 80),
                             ("octogesima", 80),
                             ("nonagésimo", 90),
                             ("nonagesimo", 90),
                             ("nonagésima", 90),
                             ("nonagesima", 90),
                             ("centesimo", 100),
                             ("centésimo", 100),
                             ("centesima", 100),
                             ("centésima", 100),
                             ("ducentésimo", 200),
                             ("ducentesimo", 200),
                             ("ducentésima", 200),
                             ("ducentesima", 200),
                             ("tricentésimo", 300),
                             ("tricentesimo", 300),
                             ("tricentésima", 300),
                             ("tricentesima", 300),
                             ("trecentésimo", 300),
                             ("trecentesimo", 300),
                             ("trecentésima", 300),
                             ("trecentesima", 300),
                             ("quadringentésimo", 400),
                             ("quadringentesimo", 400),
                             ("quadringentésima", 400),
                             ("quadringentesima", 400),
                             ("quingentésimo", 500),
                             ("quingentesimo", 500),
                             ("quingentésima", 500),
                             ("quingentesima", 500),
                             ("sexcentésimo", 600),
                             ("sexcentesimo", 600),
                             ("sexcentésima", 600),
                             ("sexcentesima", 600),
                             ("seiscentésimo", 600),
                             ("seiscentesimo", 600),
                             ("seiscentésima", 600),
                             ("seiscentesima", 600),
                             ("septingentésimo", 700),
                             ("septingentesimo", 700),
                             ("septingentésima", 700),
                             ("septingentesima", 700),
                             ("setingentésimo", 700),
                             ("setingentesimo", 700),
                             ("setingentésima", 700),
                             ("setingentesima", 700),
                             ("octingentésimo", 800),
                             ("octingentesimo", 800),
                             ("octingentésima", 800),
                             ("octingentesima", 800),
                             ("noningentésimo", 900),
                             ("noningentesimo", 900),
                             ("noningentésima", 900),
                             ("noningentesima", 900),
                             ("nongentésimo", 900),
                             ("nongentesimo", 900),
                             ("nongentésima", 900),
                             ("nongentesima", 900),
                             ("milésimo", 1000),
                             ("milesimo", 1000),
                             ("milésima", 1000),
                             ("milesima", 1000),
                             ("milionésimo", 1000000),
                             ("milionesimo", 1000000),
                             ("milionésima", 1000000),
                             ("milionesima", 1000000),
                             ("bilionésimo", 1000000000),
                             ("bilionesimo", 1000000000),
                             ("bilionésima", 1000000000),
                             ("bilionesima", 1000000000)])
    PrefixCardinalMap = dict([("hum", 1),
                              ("dois", 2),
                              ("tres", 3),
                              ("três", 3),
                              ("quatro", 4),
                              ("cinco", 5),
                              ("seis", 6),
                              ("sete", 7),
                              ("oito", 8),
                              ("nove", 9),
                              ("dez", 10),
                              ("onze", 11),
                              ("doze", 12),
                              ("treze", 13),
                              ("catorze", 14),
                              ("quatorze", 14),
                              ("quinze", 15),
                              ("dezesseis", 16),
                              ("dezasseis", 16),
                              ("dezessete", 17),
                              ("dezassete", 17),
                              ("dezoito", 18),
                              ("dezenove", 19),
                              ("dezanove", 19),
                              ("vinte", 20),
                              ("trinta", 30),
                              ("quarenta", 40),
                              ("cinquenta", 50),
                              ("cincoenta", 50),
                              ("sessenta", 60),
                              ("setenta", 70),
                              ("oitenta", 80),
                              ("noventa", 90),
                              ("cem", 100),
                              ("duzentos", 200),
                              ("trezentos", 300),
                              ("quatrocentos", 400),
                              ("quinhentos", 500),
                              ("seiscentos", 600),
                              ("setecentos", 700),
                              ("oitocentos", 800),
                              ("novecentos", 900)])
    SuffixOrdinalMap = dict([("milesimo", 1000),
                             ("milionesimo", 1000000),
                             ("bilionesimo", 1000000000),
                             ("trilionesimo", 1000000000000)])
    RoundNumberMap = dict([("mil", 1000),
                           ("milesimo", 1000),
                           ("milhão", 1000000),
                           ("milhao", 1000000),
                           ("milhões", 1000000),
                           ("milhoes", 1000000),
                           ("milionésimo", 1000000),
                           ("milionesimo", 1000000),
                           ("bilhão", 1000000000),
                           ("bilhao", 1000000000),
                           ("bilhões", 1000000000),
                           ("bilhoes", 1000000000),
                           ("bilionésimo", 1000000000),
                           ("bilionesimo", 1000000000),
                           ("trilhão", 1000000000000),
                           ("trilhao", 1000000000000),
                           ("trilhões", 1000000000000),
                           ("trilhoes", 1000000000000),
                           ("trilionésimo", 1000000000000),
                           ("trilionesimo", 1000000000000),
                           ("dezena", 10),
                           ("dezenas", 10),
                           ("dúzia", 12),
                           ("duzia", 12),
                           ("dúzias", 12),
                           ("duzias", 12),
                           ("k", 1000),
                           ("m", 1000000),
                           ("g", 1000000000),
                           ("b", 1000000000),
                           ("t", 1000000000000)])
    AmbiguityFiltersDict = dict([("^[.]", "")])
    RelativeReferenceOffsetMap = dict([("", "")])
    RelativeReferenceRelativeToMap = dict([("", "")])
# pylint: enable=line-too-long



from enum import Enum

from recognizers_number.number.parsers import NumberParserConfiguration, BaseNumberParser, BasePercentageParser
from recognizers_number.number.constants import Constants
from recognizers_number.number.cjk_parsers import CJKNumberParser
from recognizers_number.number.chinese.parsers import ChineseNumberParserConfiguration
from recognizers_number.number.japanese.parsers import JapaneseNumberParserConfiguration


class ParserType(Enum):
    NUMBER = 0
    CARDINAL = 1
    DOUBLE = 2
    FRACTION = 3
    INTEGER = 4
    ORDINAL = 5
    PERCENTAGE = 6


class AgnosticNumberParserFactory:
    @staticmethod
    def get_parser(parser_type: ParserType, language_config: NumberParserConfiguration) -> BaseNumberParser:
        parser = BaseNumberParser(language_config)

        chinese = isinstance(language_config, ChineseNumberParserConfiguration)
        japanese = isinstance(
            language_config, JapaneseNumberParserConfiguration)

        if chinese:
            parser = CJKNumberParser(language_config)
        elif japanese:
            parser = CJKNumberParser(language_config)

        if parser_type is ParserType.CARDINAL:
            parser.supported_types = [
                Constants.SYS_NUM_CARDINAL,
                Constants.SYS_NUM_INTEGER,
                Constants.SYS_NUM_DOUBLE
            ]
        elif parser_type is ParserType.DOUBLE:
            parser.supported_types = [Constants.SYS_NUM_DOUBLE]
        elif parser_type is ParserType.FRACTION:
            parser.supported_types = [Constants.SYS_NUM_FRACTION]
        elif parser_type is ParserType.INTEGER:
            parser.supported_types = [Constants.SYS_NUM_INTEGER]
        elif parser_type is ParserType.ORDINAL:
            parser.supported_types = [Constants.SYS_NUM_ORDINAL]
        elif parser_type is ParserType.PERCENTAGE and not chinese:
            parser = BasePercentageParser(language_config)

        return parser



from typing import List
from enum import Enum

from recognizers_number.number.extractors import ReVal, BaseNumberExtractor
from recognizers_text.utilities import RegExpUtility
from recognizers_number.number.constants import Constants
from recognizers_number.resources.chinese_numeric import ChineseNumeric


class ChineseNumberExtractorMode(Enum):
    DEFAULT = 0
    EXTRACT_ALL = 1


class ChineseNumberExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[ReVal]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM

    def __init__(self, mode: ChineseNumberExtractorMode = ChineseNumberExtractorMode.DEFAULT):
        self.__regexes: List[ReVal] = list()

        cardinal_ex = ChineseCardinalExtractor(mode)
        self.__regexes.extend(cardinal_ex.regexes)

        fraction_ex = ChineseFractionExtractor()
        self.__regexes.extend(fraction_ex.regexes)


class ChineseCardinalExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[ReVal]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_CARDINAL

    def __init__(self, mode: ChineseNumberExtractorMode = ChineseNumberExtractorMode.DEFAULT):
        self.__regexes: List[ReVal] = list()

        integer_ex = ChineseIntegerExtractor(mode)
        self.__regexes.extend(integer_ex.regexes)

        double_ex = ChineseDoubleExtractor()
        self.__regexes.extend(double_ex.regexes)


class ChineseIntegerExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[ReVal]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_INTEGER

    def __init__(self, mode: ChineseNumberExtractorMode = ChineseNumberExtractorMode.DEFAULT):
        self.__regexes = [
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.NumbersSpecialsChars),
                val='IntegerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.NumbersSpecialsCharsWithSuffix),
                val='IntegerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.DottedNumbersSpecialsChar),
                val='IntegerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.NumbersWithHalfDozen),
                val='IntegerChs'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.NumbersWithDozen),
                val='IntegerChs'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.HalfUnitRegex),
                val='IntegerChs')
        ]
        if mode == ChineseNumberExtractorMode.DEFAULT:
            self.__regexes.append(
                ReVal(
                    re=RegExpUtility.get_safe_reg_exp(
                        ChineseNumeric.NumbersWithAllowListRegex),
                    val='IntegerChs'
                )
            )
        elif mode == ChineseNumberExtractorMode.EXTRACT_ALL:
            self.__regexes.append(
                ReVal(
                    re=RegExpUtility.get_safe_reg_exp(
                        ChineseNumeric.NumbersAggressiveRegex),
                    val='IntegerChs'
                )
            )


class ChineseDoubleExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[ReVal]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_DOUBLE

    def __init__(self):
        self.__regexes = [
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.DoubleSpecialsChars),
                val='DoubleNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.DoubleSpecialsCharsWithNegatives),
                val='DoubleNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.SimpleDoubleSpecialsChars),
                val='DoubleNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.DoubleWithMultiplierRegex),
                val='DoubleNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.DoubleWithThousandsRegex),
                val='DoubleChs'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.DoubleAllFloatRegex),
                val='DoubleChs'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.DoubleExponentialNotationRegex),
                val='DoublePow'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.DoubleScientificNotationRegex),
                val='DoublePow')
        ]


class ChineseFractionExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[ReVal]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_FRACTION

    def __init__(self):
        self.__regexes = [
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.FractionNotationSpecialsCharsRegex),
                val='FracNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.FractionNotationRegex),
                val='FracNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.AllFractionNumber),
                val='FracChs')
        ]


class ChineseOrdinalExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[ReVal]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_ORDINAL

    def __init__(self):
        self.__regexes = [
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(ChineseNumeric.OrdinalRegex),
                val='OrdinalChs'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.OrdinalNumbersRegex),
                val='OrdinalChs')
        ]


class ChinesePercentageExtractor(BaseNumberExtractor):
    @property
    def regexes(self) -> List[ReVal]:
        return self.__regexes

    @property
    def _extract_type(self) -> str:
        return Constants.SYS_NUM_PERCENTAGE

    def __init__(self):
        self.__regexes = [
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.PercentagePointRegex),
                val='PerChs'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.SimplePercentageRegex),
                val='PerChs'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.NumbersPercentagePointRegex),
                val='PerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.NumbersPercentageWithSeparatorRegex),
                val='PerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.NumbersPercentageWithMultiplierRegex),
                val='PerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.FractionPercentagePointRegex),
                val='PerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.FractionPercentageWithSeparatorRegex),
                val='PerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.FractionPercentageWithMultiplierRegex),
                val='PerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.SimpleNumbersPercentageRegex),
                val='PerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.SimpleNumbersPercentageWithMultiplierRegex),
                val='PerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.SimpleNumbersPercentagePointRegex),
                val='PerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.IntegerPercentageRegex),
                val='PerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.IntegerPercentageWithMultiplierRegex),
                val='PerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.NumbersFractionPercentageRegex),
                val='PerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.SimpleIntegerPercentageRegex),
                val='PerNum'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.NumbersFoldsPercentageRegex),
                val='PerSpe'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.FoldsPercentageRegex),
                val='PerSpe'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.SimpleFoldsPercentageRegex),
                val='PerSpe'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.SpecialsPercentageRegex),
                val='PerSpe'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.NumbersSpecialsPercentageRegex),
                val='PerSpe'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.SimpleSpecialsPercentageRegex),
                val='PerSpe'),
            ReVal(
                re=RegExpUtility.get_safe_reg_exp(
                    ChineseNumeric.SpecialsFoldsPercentageRegex),
                val='PerSpe')
        ]



from abc import abstractmethod
from typing import List, Dict, Pattern, Optional
from collections import namedtuple
from decimal import Decimal, getcontext
import copy
import regex

from recognizers_text.utilities import RegExpUtility
from recognizers_text.culture import Culture
from recognizers_text.extractor import ExtractResult
from recognizers_text.parser import ParseResult
from recognizers_number.number.parsers import BaseNumberParser, NumberParserConfiguration
from recognizers_number.culture import CultureInfo

getcontext().prec = 15


class CJKNumberParserConfiguration(NumberParserConfiguration):
    @property
    @abstractmethod
    def zero_to_nine_map(self) -> Dict[str, int]:
        pass

    @property
    @abstractmethod
    def round_number_map_char(self) -> Dict[str, int]:
        pass

    @property
    @abstractmethod
    def full_to_half_map(self) -> Dict[str, int]:
        pass

    @property
    @abstractmethod
    def trato_sim_map(self) -> Dict[str, int]:
        pass

    @property
    @abstractmethod
    def unit_map(self) -> Dict[str, int]:
        pass

    @property
    @abstractmethod
    def round_direct_list(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def ten_chars(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def digit_num_regex(self) -> Pattern:
        pass

    @property
    @abstractmethod
    def dozen_regex(self) -> Pattern:
        pass

    @property
    @abstractmethod
    def percentage_regex(self) -> Pattern:
        pass

    @property
    @abstractmethod
    def double_and_round_regex(self) -> Pattern:
        pass

    @property
    @abstractmethod
    def frac_split_regex(self) -> Pattern:
        pass

    @property
    @abstractmethod
    def point_regex(self) -> Pattern:
        pass

    @property
    @abstractmethod
    def spe_get_number_regex(self) -> Pattern:
        pass

    @property
    @abstractmethod
    def pair_regex(self) -> Pattern:
        pass

    @property
    @abstractmethod
    def round_number_integer_regex(self) -> Pattern:
        pass

    @property
    @abstractmethod
    def zero_char(self) -> str:
        pass

    @property
    @abstractmethod
    def pair_char(self) -> str:
        pass


class CJKNumberParser(BaseNumberParser):
    def __init__(self, config: CJKNumberParserConfiguration):
        super().__init__(config)
        self.config = config

    def __format(self, value: object) -> str:
        if self.config.culture_info is None:
            return str(value)
        return self.config.culture_info.format(value)

    def parse(self, source: ExtractResult) -> Optional[ParseResult]:
        result: ParseResult
        extra: str = source.data
        simplified_source: ExtractResult = copy.deepcopy(source)
        simplified_source.text = self.replace_trad_with_simplified(source.text)

        if not extra:
            return result

        if 'Per' in extra:
            result = self.per_parse(simplified_source)
        elif 'Num' in extra:
            simplified_source.text = self.replace_full_with_half(
                simplified_source.text)
            result = self._digit_number_parse(simplified_source)
            if regex.search(self.config.negative_number_sign_regex, simplified_source.text) and result.value > 0:
                result.value = result.value * -1
            result.resolution_str = self.__format(result.value)
        elif 'Pow' in extra:
            simplified_source.text = self.replace_full_with_half(
                simplified_source.text)
            result = self._power_number_parse(simplified_source)
            result.resolution_str = self.__format(result.value)
        elif 'Frac' in extra:
            result = self.frac_parse(simplified_source)
        elif 'Dou' in extra:
            result = self.dou_parse(simplified_source)
        elif 'Integer' in extra:
            result = self.int_parse(simplified_source)
        elif 'Ordinal' in extra:
            result = self.ord_parse(simplified_source)

        if result is not None:
            result.text = source.text.lower()

        return result

    def replace_trad_with_simplified(self, source: str) -> str:
        if source is None or not source.strip():
            return source
        if self.config.trato_sim_map is None:
            return source
        return ''.join(map(lambda c: self.config.trato_sim_map.get(c, c), source))

    def replace_full_with_half(self, source: str) -> str:
        if source is None or not source.strip():
            return source
        return ''.join(map(lambda c: self.config.full_to_half_map.get(c, c), source))

    def replace_unit(self, source: str) -> str:
        if source is None or not source.strip():
            return source
        for (k, v) in self.config.unit_map.items():
            source = source.replace(k, v)
        return source

    def per_parse(self, source: ExtractResult) -> ParseResult:
        result = ParseResult(source)
        source_text = source.text
        power = 1

        if 'Spe' in source.data:
            source_text = self.replace_full_with_half(source_text)
            source_text = self.replace_unit(source_text)

            if source_text == '半額' or source_text == '半折':
                result.value = 50
            elif source_text == '10成' or source_text == '10割' or source_text == '十割':
                result.value = 100
            else:
                matches = list(regex.finditer(
                    self.config.spe_get_number_regex, source_text))
                int_number: int
                if len(matches) == 2:
                    int_number_char = matches[0].group()[0]
                    if int_number_char == self.config.pair_char:
                        int_number = 5
                    elif int_number_char in self.config.ten_chars:
                        int_number = 10
                    else:
                        int_number = self.config.zero_to_nine_map[int_number_char]

                    point_number_char = matches[1].group()[0]
                    point_number: float
                    if point_number_char == '半':
                        point_number = 0.5
                    else:
                        point_number = self.config.zero_to_nine_map[point_number_char] * 0.1

                    result.value = (int_number + point_number) * 10
                elif len(matches) == 5:
                    # Deal the Japanese percentage case like "xxx割xxx分xxx厘", get the integer value and convert into result.
                    int_number_char = matches[0].group()[0]
                    point_number_char = matches[1].group()[0]
                    dot_number_char = matches[3].group()[0]

                    point_number = self.config.zero_to_nine_map[point_number_char] * 0.1
                    dot_number = self.config.zero_to_nine_map[dot_number_char] * 0.01

                    int_number = self.config.zero_to_nine_map[int_number_char]

                    result.value = (
                        int_number + point_number + dot_number) * 10
                else:
                    int_number_char = matches[0].group()[0]
                    if int_number_char == self.config.pair_char:
                        int_number = 5
                    elif int_number_char in self.config.ten_chars:
                        int_number = 10
                    else:
                        int_number = self.config.zero_to_nine_map[int_number_char]
                    result.value = int_number * 10

        elif 'Num' in source.data:
            double_match = regex.search(
                self.config.percentage_regex, source_text)
            double_text = double_match.group()

            if any(x for x in ['k', 'K', 'ｋ', 'Ｋ'] if x in double_text):
                power = 1000
            elif any(x for x in ['M', 'Ｍ'] if x in double_text):
                power = 1000000
            elif any(x for x in ['G', 'Ｇ'] if x in double_text):
                power = 1000000000
            elif any(x for x in ['T', 'Ｔ'] if x in double_text):
                power = 1000000000000
            result.value = self.get_digit_value(double_text, power)

        else:
            double_match = regex.search(
                self.config.percentage_regex, source_text)
            double_text = self.replace_unit(double_match.group())

            split_result = regex.split(self.config.point_regex, double_text)
            if split_result[0] == '':
                split_result[0] = self.config.zero_char

            double_value = self.get_int_value(split_result[0])
            if len(split_result) == 2:
                if regex.search(self.config.negative_number_sign_regex, split_result[0]) is not None:
                    double_value -= self.get_point_value(split_result[1])
                else:
                    double_value += self.get_point_value(split_result[1])
            result.value = double_value

        result.resolution_str = self.__format(result.value) + '%'
        return result

    def frac_parse(self, source: ExtractResult) -> ParseResult:
        result = ParseResult(source)

        source_text = source.text
        split_result = regex.split(self.config.frac_split_regex, source_text)

        parts = namedtuple('parts', ['intval', 'demo', 'num'])

        result_part: parts

        if len(split_result) == 3:
            result_part = parts(
                intval=split_result[0],
                demo=split_result[1],
                num=split_result[2]
            )
        else:
            result_part = parts(
                intval=self.config.zero_char,
                demo=split_result[0],
                num=split_result[1]
            )

        int_value = Decimal(self.get_value_from_part(result_part.intval))
        num_value = Decimal(self.get_value_from_part(result_part.num))
        demo_value = Decimal(self.get_value_from_part(result_part.demo))

        if regex.search(self.config.negative_number_sign_regex, result_part.intval) is not None:
            result.value = int_value - num_value / demo_value
        else:
            result.value = int_value + num_value / demo_value

        result.resolution_str = self.__format(result.value)
        return result

    def get_value_from_part(self, part: str) -> float:
        if self.is_digit(part):
            return self.get_digit_value(part, 1.0)
        return self.get_int_value(part)

    def dou_parse(self, source: ExtractResult) -> ParseResult:
        result = ParseResult(source)

        source_text = self.replace_unit(source.text)

        if (regex.search(self.config.double_and_round_regex, source.text)) is not None:
            power = self.config.round_number_map_char[source_text[-1:]]
            result.value = self.get_digit_value(source_text[:-1], power)
        else:
            split_result = regex.split(self.config.point_regex, source_text)
            if split_result[0] == '':
                split_result[0] = self.config.zero_char
            if regex.search(self.config.negative_number_sign_regex, split_result[0]) is not None:
                result.value = self.get_int_value(
                    split_result[0]) - self.get_point_value(split_result[1])
            else:
                result.value = self.get_int_value(
                    split_result[0]) + self.get_point_value(split_result[1])

        result.resolution_str = self.__format(result.value)
        return result

    def int_parse(self, source: ExtractResult) -> ParseResult:
        result = ParseResult(source)
        result.value = self.get_int_value(source.text)
        result.resolution_str = self.__format(result.value)
        return result

    def ord_parse(self, source: ExtractResult) -> ParseResult:
        result = ParseResult(source)
        source_text = source.text[1:]

        if regex.search(self.config.digit_num_regex, source_text) is not None:
            result.value = self.get_digit_value(source_text, 1)
        else:
            result.value = self.get_int_value(source_text)

        result.resolution_str = self.__format(result.value)
        return result

    def get_digit_value(self, source: str, power: float) -> float:
        negative: bool = False
        result_str = source
        if regex.search(self.config.negative_number_sign_regex, result_str) is not None:
            negative = True
            result_str = result_str[1:]
        result_str = self.replace_full_with_half(result_str)
        result = float(super()._get_digital_value(result_str, power))
        if negative:
            result = - result
        return result

    def get_int_value(self, source: str) -> float:
        result_str = source
        dozen = False
        pair = False

        if regex.search(self.config.dozen_regex, result_str) is not None:
            dozen = True
            if self.config.culture_info.code == Culture.Chinese:
                #            if isinstance(self.config, ChineseNumberParserConfiguration):
                result_str = result_str[:-1]
            elif self.config.culture_info.code == Culture.Japanese:
                #            if isinstance(self.config, CJKNumberParserConfiguration):
                result_str = result_str[:-3]
        elif regex.search(self.config.pair_regex, result_str) is not None:
            pair = True
            result_str = result_str[:-1]

        result_str = self.replace_unit(result_str)
        int_value = 0
        part_value = 0
        before_value = 1
        is_round_before = False
        round_before = -1
        round_default = 1
        negative = False

        if regex.search(self.config.negative_number_sign_regex, result_str) is not None:
            negative = True
            result_str = result_str[1:]

        for i in range(len(result_str)):
            c = result_str[i]
            if c in self.config.round_number_map_char:
                round_recent = self.config.round_number_map_char[c]
                if round_before != -1 and round_recent > round_before:
                    if is_round_before:
                        int_value += part_value * round_recent
                        is_round_before = False
                    else:
                        part_value += before_value * round_default
                        int_value += part_value * round_recent
                    round_before = -1
                    part_value = 0
                else:
                    is_round_before = True
                    part_value += before_value * round_recent
                    round_before = round_recent
                    if i == len(result_str)-1 or c in self.config.round_direct_list:
                        int_value += part_value
                        part_value = 0

                round_default = round_recent / 10
            elif c in self.config.zero_to_nine_map:
                if i != len(result_str)-1:
                    is_not_round_next = result_str[i + 1] in self.config.ten_chars or result_str[
                        i + 1] not in self.config.round_number_map_char
                    if c == self.config.zero_char and is_not_round_next:
                        before_value = 1
                        round_default = 1
                    else:
                        before_value = self.config.zero_to_nine_map[c]
                        is_round_before = False
                else:
                    if i == len(result_str)-1 and self.config.culture_info.code == Culture.Japanese:
                        round_default = 1
                    part_value += self.config.zero_to_nine_map[c] * \
                        round_default
                    int_value += part_value
                    part_value = 0
        if negative:
            int_value = - int_value
        if dozen:
            int_value = int_value * 12
        if pair:
            int_value = int_value * 2

        return int_value

    def get_point_value(self, source: str) -> float:
        result: float = 0
        scale: float = 0.1
        for c in source:
            result += scale * self.config.zero_to_nine_map[c]
            scale *= 0.1
        return result

    def is_digit(self, source: str) -> bool:
        return (source is not None
                and len(source.strip()) > 0
                and regex.search(self.config.digit_num_regex, source) is not None)



from abc import ABC, abstractmethod
from typing import Dict, Pattern, Optional, List
from decimal import Decimal, getcontext
import regex
from recognizers_text.utilities import RegExpUtility
from recognizers_text.extractor import ExtractResult
from recognizers_text.parser import Parser, ParseResult
from recognizers_number.culture import CultureInfo
from recognizers_number.number.constants import Constants

getcontext().prec = 15


class NumberParserConfiguration(ABC):
    @property
    @abstractmethod
    def cardinal_number_map(self) -> Dict[str, int]:
        pass

    @property
    @abstractmethod
    def ordinal_number_map(self) -> Dict[str, int]:
        pass

    @property
    @abstractmethod
    def round_number_map(self) -> Dict[str, int]:
        pass

    @property
    @abstractmethod
    def culture_info(self) -> CultureInfo:
        pass

    @property
    @abstractmethod
    def digital_number_regex(self) -> Pattern:
        pass

    @property
    @abstractmethod
    def fraction_marker_token(self) -> str:
        pass

    @property
    @abstractmethod
    def negative_number_sign_regex(self) -> Pattern:
        pass

    @property
    @abstractmethod
    def half_a_dozen_regex(self) -> Pattern:
        pass

    @property
    @abstractmethod
    def half_a_dozen_text(self) -> str:
        pass

    @property
    @abstractmethod
    def lang_marker(self) -> str:
        pass

    @property
    @abstractmethod
    def non_decimal_separator_char(self) -> str:
        pass

    @property
    @abstractmethod
    def decimal_separator_char(self) -> str:
        pass

    @property
    @abstractmethod
    def word_separator_token(self) -> str:
        pass

    @property
    @abstractmethod
    def written_decimal_separator_texts(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def written_group_separator_texts(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def written_integer_separator_texts(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def written_fraction_separator_texts(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def normalize_token_set(self, tokens: List[str], context: ParseResult) -> List[str]:
        pass

    @property
    @abstractmethod
    def resolve_composite_number(self, number_str: str) -> int:
        pass


class BaseNumberParser(Parser):
    def __init__(self, config: NumberParserConfiguration):
        self.config: NumberParserConfiguration = config
        self.supported_types: List[str] = list()

        single_int_frac = f'{self.config.word_separator_token}| -|{self._get_key_regex(self.config.cardinal_number_map.keys())}|{self._get_key_regex(self.config.ordinal_number_map.keys())}'
        self.text_number_regex: Pattern = RegExpUtility.get_safe_reg_exp(
            fr'(?=\b)({single_int_frac})(?=\b)', flags=regex.I | regex.S)
        self.arabic_number_regex: Pattern = RegExpUtility.get_safe_reg_exp(
            r'\d+', flags=regex.I | regex.S)
        self.round_number_set: List[str] = list(
            self.config.round_number_map.keys())

    def parse(self, source: ExtractResult) -> Optional[ParseResult]:
        # Check if the parser is configured to support specific types
        if self.supported_types and source.type not in self.supported_types:
            return None
        ret: Optional[ParseResult] = None
        extra = source.data if isinstance(source.data, str) else None
        if not extra:
            if self.arabic_number_regex.search(source.text):
                extra = 'Num'
            else:
                extra = self.config.lang_marker

        # Resolve symbol prefix
        is_negative = False
        match_negative = regex.search(
            self.config.negative_number_sign_regex, source.text)

        if match_negative:
            is_negative = True
            source.text = source.text[len(match_negative[1]):]

        if 'Num' in extra:
            ret = self._digit_number_parse(source)
        # Frac is a special number, parse via another method
        elif regex.search(fr'Frac{self.config.lang_marker}', extra):
            ret = self._frac_like_number_parse(source)
        elif self.config.lang_marker in extra:
            ret = self._text_number_parse(source)
        elif 'Pow' in extra:
            ret = self._power_number_parse(source)

        if ret and ret.value is not None:
            if is_negative:
                # Recover to the original extracted Text
                ret.text = match_negative[1] + source.text
                ret.value = ret.value * -1
            # Use culture_info to format values
            ret.resolution_str = self.config.culture_info.format(
                ret.value) if self.config.culture_info is not None else repr(ret.value)
            ret.text = ret.text.lower()

        return ret

    def _get_key_regex(self, keys: List[str]) -> str:
        return str.join('|', sorted(keys, key=len, reverse=True))

    def _digit_number_parse(self, ext_result: ExtractResult) -> ParseResult:
        result = ParseResult()
        result.start = ext_result.start
        result.length = ext_result.length
        result.text = ext_result.text
        result.type = ext_result.type

        # [1] 24
        # [2] 12 32/33
        # [3] 1,000,000
        # [4] 234.567
        # [5] 44/55
        # [6] 2 hundred
        # dot occurred.

        power = 1
        tmp_index = -1
        start_index = 0
        handle = ext_result.text.lower()

        matches = list(regex.finditer(
            self.config.digital_number_regex, handle))
        if matches:
            for match in matches:
                rep = self.config.round_number_map.get(match.group())
                # \\s+ for filter the spaces.
                power *= rep

                tmp_index = handle.find(match.group(), start_index)
                while tmp_index >= 0:
                    front = handle[0:tmp_index].rstrip()
                    start_index = len(front)
                    handle = front + handle[tmp_index + len(match):]
                    tmp_index = handle.find(match.group(), start_index)

        # Scale used in the calculate of double
        result.value = self._get_digital_value(handle, power)

        return result

    @staticmethod
    def _is_digit(c: str) -> bool:
        return c.isdigit()

    def _frac_like_number_parse(self, ext_result: ExtractResult) -> ParseResult:
        result = ParseResult()
        result.start = ext_result.start
        result.length = ext_result.length
        result.text = ext_result.text
        result.type = ext_result.type

        result_text = ext_result.text.lower()
        if regex.search(self.config.fraction_marker_token, result_text):
            over_index = result_text.find(self.config.fraction_marker_token)
            small_part = result_text[0:over_index].strip()
            big_part = result_text[over_index +
                                   len(self.config.fraction_marker_token):len(result_text)].strip()
            small_value = self._get_digital_value(small_part, 1) if self._is_digit(
                small_part[0]) else self.__get_int_value(self.__get_matches(small_part))
            big_value = self._get_digital_value(big_part, 1) if self._is_digit(
                big_part[0]) else self.__get_int_value(self.__get_matches(big_part))

            result.value = small_value / big_value
        else:
            words = list(filter(lambda x: x, result_text.split(' ')))
            frac_words = self.config.normalize_token_set(words, result)

            # Split fraction with integer
            split_index = len(frac_words) - 1
            current_value = self.config.resolve_composite_number(
                frac_words[split_index])
            round_value = 1

            # for case like "half"
            if len(frac_words) == 1:
                result.value = 1 / self.__get_int_value(frac_words)
                return result

            for split_index in range(len(frac_words) - 2, -1, -1):
                if (frac_words[split_index] in self.config.written_fraction_separator_texts
                        or frac_words[split_index] in self.config.written_integer_separator_texts):
                    continue
                previous_value = current_value
                current_value = self.config.resolve_composite_number(
                    frac_words[split_index])

                sm_hundreds = 100

                # previous: hundred
                # current: one
                if ((previous_value >= sm_hundreds and previous_value > current_value)
                        or (previous_value < sm_hundreds and self.__is_composable(current_value, previous_value))):
                    if (previous_value < sm_hundreds and current_value >= round_value):
                        round_value = current_value
                    elif (previous_value < sm_hundreds and current_value < round_value):
                        split_index += 1
                        break

                    # current is the first word
                    if split_index == 0:
                        # scan, skip the first word
                        split_index = 1
                        while split_index <= len(frac_words) - 2:
                            # e.g. one hundred thousand
                            # frac[i+1] % 100 and frac[i] % 100 = 0
                            if (self.config.resolve_composite_number(frac_words[split_index]) >= sm_hundreds
                                    and not frac_words[split_index + 1] in self.config.written_fraction_separator_texts
                                    and self.config.resolve_composite_number(frac_words[split_index + 1]) < sm_hundreds):
                                split_index += 1
                                break
                            split_index += 1
                        break
                    continue
                split_index += 1
                break

            frac_part = []
            for i in range(split_index, len(frac_words)):
                if frac_words[i].find('-') > -1:
                    split = frac_words[i].split('-')
                    frac_part.append(split[0])
                    frac_part.append('-')
                    frac_part.append(split[1])
                else:
                    frac_part.append(frac_words[i])

            frac_words = frac_words[:split_index]

            # denomi = denominator
            denomi_value = self.__get_int_value(frac_part)
            # Split mixed number with fraction
            numer_value = 0
            int_value = 0

            mixed_index = len(frac_words)
            for i in range(len(frac_words) - 1, -1, -1):
                if (i < len(frac_words) - 1 and frac_words[i] in self.config.written_fraction_separator_texts):
                    numer_str = ' '.join(frac_words[i + 1:len(frac_words)])
                    numer_value = self.__get_int_value(
                        self.__get_matches(numer_str))
                    mixed_index = i + 1
                    break

            int_str = ' '.join(frac_words[0:mixed_index])
            int_value = self.__get_int_value(self.__get_matches(int_str))

            # Find mixed number
            if (mixed_index != len(frac_words) and numer_value < denomi_value):
                # int_value + numer_value / denomi_value
                result.value = int_value + numer_value / denomi_value
            else:
                # (int_value + numer_value) / denomi_value
                result.value = (int_value + numer_value) / denomi_value

            # Convert to float for fixed float point vs. exponential notation consistency /w C#/TS/JS
            result.value = float(result.value)
        return result

    def _text_number_parse(self, ext_result: ExtractResult) -> ParseResult:
        result = ParseResult(ext_result)

        handle = regex.sub(self.config.half_a_dozen_regex,
                           self.config.half_a_dozen_text, ext_result.text.lower())
        num_group = self.__split_multi(handle, list(
            filter(lambda x: x is not None, self.config.written_decimal_separator_texts)))

        int_part = num_group[0]

        matches = list(map(lambda x: x.group().lower(), list(regex.finditer(
            self.text_number_regex, int_part)))) if int_part else list()

        int_part_real = self.__get_int_value(matches)

        point_part_real = Decimal(0)
        if len(num_group) == 2:
            point_part = num_group[1]
            matches = list(map(lambda x: x.group().lower(), list(
                regex.finditer(self.text_number_regex, point_part))))
            point_part_real += self.__get_point_value(matches)

        result.value = int_part_real + Decimal(point_part_real)
        return result

    def _power_number_parse(self, ext_result: ExtractResult) -> ParseResult:
        result = ParseResult(ext_result)

        handle = ext_result.text.upper()
        exponent = '^' not in ext_result.text

        # [1] 1e10
        # [2] 1.1^-23
        call_stack = list()
        scale = 10
        dot = False
        negative = False
        tmp = 0

        for i in range(len(handle)):
            c = handle[i]
            if c in ['^', 'E']:
                if negative:
                    call_stack.append(-tmp)
                else:
                    call_stack.append(tmp)
                tmp = 0
                scale = 10
                dot = False
                negative = False
            elif c.isdigit():
                if dot:
                    tmp = tmp + scale * int(c)
                    scale *= 0.1
                else:
                    tmp = tmp * scale + int(c)
            elif c == self.config.decimal_separator_char:
                dot = True
                scale = 0.1
            elif c == '-':
                negative = not negative
            elif c == '+':
                continue
            if i == len(handle)-1:
                if negative:
                    call_stack.append(-tmp)
                else:
                    call_stack.append(tmp)
        result_value = 0
        a = Decimal(call_stack.pop(0))
        b = Decimal(call_stack.pop(0))
        if exponent:
            result_value = getcontext().multiply(a, getcontext().power(Decimal(10), b))
        else:
            result_value = getcontext().power(a, b)

        result.value = result_value
        result.resolution_str = str(result_value)

        return result

    def __split_multi(self, source: str, tokens: List[str]) -> List[str]:
        # We can use the first token as a temporary join character
        tmp = tokens[0]
        for token in tokens:
            source = tmp.join(source.split(token))
        return source.split(tmp)

    def __get_matches(self, source: str) -> List[str]:
        matches = list(regex.finditer(self.text_number_regex, source))
        return list(filter(None, map(lambda m: m.group().lower(), matches)))

    # Test if big and combine with small.
    # e.g. 'hundred' can combine with 'thirty' but 'twenty' can't combine with 'thirty'.
    def __is_composable(self, big: int, small: int) -> bool:
        base_num = 100 if small > 10 else 10
        return big % base_num == 0 and big / base_num >= 1

    def __get_int_value(self, matches: List[str]) -> Decimal:
        is_end = [False] * len(matches)

        tmp_val = 0
        end_flag = 1

        # Scan from end to start, find the end word
        for i in range(len(matches)-1, 0, -1):
            if matches[i] in self.round_number_set:
                # if false,then continue, you will meet hundred first, then thousand.
                if end_flag > self.config.round_number_map[matches[i]]:
                    continue
                is_end[i] = True
                end_flag = self.config.round_number_map[matches[i]]

        if end_flag == 1:
            tmp_stack = list()
            old_sym = ''
            for match in matches:
                cardinal = match in self.config.cardinal_number_map
                ordinal = match in self.config.ordinal_number_map
                if cardinal or ordinal:
                    match_value = self.config.cardinal_number_map.get(
                        match, None)
                    if match_value is None:
                        match_value = self.config.ordinal_number_map.get(
                            match, None)

                    # This is just for ordinal now. Not for fraction ever.
                    if ordinal:
                        frac_part = self.config.ordinal_number_map[match]
                        if tmp_stack:
                            int_part = tmp_stack.pop()
                            # if intPart >= frac_part, it means it is an ordinal number
                            # it begins with an integer, ends with an ordinal
                            # e.g. ninety-ninth
                            if int_part >= frac_part:
                                tmp_stack.append(int_part + frac_part)
                            else:
                                for val in tmp_stack:
                                    int_part += val
                                tmp_stack.clear()
                                tmp_stack.append(int_part * frac_part)
                        else:
                            tmp_stack.append(frac_part)
                    elif match in self.config.cardinal_number_map:
                        if old_sym == '-':
                            tmp = tmp_stack.pop() + match_value
                            tmp_stack.append(tmp)
                        elif old_sym == self.config.written_integer_separator_texts[0] or len(tmp_stack) < 2:
                            tmp_stack.append(match_value)
                        elif len(tmp_stack) >= 2:
                            tmp = tmp_stack.pop() + match_value
                            tmp += tmp_stack.pop()
                            tmp_stack.append(tmp)
                else:
                    complex_val = self.config.resolve_composite_number(match)
                    if complex_val != 0:
                        tmp_stack.append(complex_val)
                old_sym = match
            for tmp in tmp_stack:
                tmp_val += tmp
        else:
            last_index = 0
            part_value = 1
            mul_value = 1
            for i in range(len(is_end)):
                if is_end[i]:
                    mul_value = self.config.round_number_map[matches[i]]
                    part_value = 1
                    if i != 0:
                        part_value = self.__get_int_value(
                            matches[last_index:i])
                    tmp_val += mul_value * part_value
                    last_index = i + 1
            # Calculate the part like 'thirty-one'
            mul_value = 1
            if last_index != len(is_end):
                part_value = self.__get_int_value(
                    matches[last_index:len(is_end)])
                tmp_val += mul_value * part_value
        return Decimal(tmp_val)

    def __get_point_value(self, matches: List[str]) -> Decimal:
        result = 0
        first_match = matches[0]
        if first_match in self.config.cardinal_number_map and self.config.cardinal_number_map[first_match] >= 10:
            prefix = '0.'
            tmp_int = self.__get_int_value(matches)
            result = Decimal(prefix + str(tmp_int))
        else:
            scale = Decimal(0.1)
            for match in matches:
                result += scale * \
                    Decimal(self.config.cardinal_number_map[match])
                scale *= Decimal(0.1)

        return result

    def __skip_non_decimal_separator(self, ch: str, distance: int, culture: CultureInfo) -> bool:

        decimal_length: int = 3

        # Special cases for multi-language countries where decimal separators can be used interchangeably. Mostly informally.
        # Ex: South Africa, Namibia; Puerto Rico in ES; or in Canada for EN and FR.
        # "me pidio $5.00 prestados" and "me pidio $5,00 prestados" -> currency $5
        culture_regex: Pattern = RegExpUtility.get_safe_reg_exp(
            r'^(en|es|fr)(-)?\b', flags=regex.I | regex.S)

        return ch == self.config.non_decimal_separator_char and not(distance <= decimal_length and culture_regex.match(culture.code))

    def _get_digital_value(self, digits_str: str, power: int) -> Decimal:
        tmp: Decimal = Decimal(0)
        scale: Decimal = Decimal(10)
        decimal_separator: bool = False
        str_length: int = len(digits_str)
        negative: bool = False
        fraction: bool = '/' in digits_str
        index: int = 0

        call_stack: List[Decimal] = list()

        for c in digits_str:

            skippable_non_decimal = self.__skip_non_decimal_separator(
                c, str_length - index, self.config.culture_info)
            index += 1

            if not fraction and (c == ' ' or c == Constants.NO_BREAK_SPACE or skippable_non_decimal):
                continue

            if c == ' ' or c == '/':
                call_stack.append(tmp)
                tmp = Decimal(0)
            elif c.isdigit():
                if decimal_separator:
                    tmp = getcontext().add(tmp, getcontext().multiply(scale, Decimal(c)))
                    scale = getcontext().multiply(scale, Decimal(0.1))
                else:
                    tmp = getcontext().add(getcontext().multiply(tmp, scale), Decimal(c))
            elif c == self.config.decimal_separator_char or (not skippable_non_decimal and c == self.config.non_decimal_separator_char):
                decimal_separator = True
                scale = Decimal(0.1)
            elif c == '-':
                negative = True

        call_stack.append(tmp)

        # is the number is a fraction.
        cal_result = Decimal(0)
        if fraction:
            deno = call_stack.pop()
            mole = call_stack.pop()
            cal_result = getcontext().add(cal_result, getcontext().divide(mole, deno))

        for n in call_stack:
            cal_result = getcontext().add(cal_result, n)

        cal_result = getcontext().multiply(cal_result, Decimal(power))

        return cal_result if not negative else cal_result * -1


class BasePercentageParser(BaseNumberParser):
    def parse(self, source: ExtractResult) -> Optional[ParseResult]:
        original = source.text

        # do replace text & data from extended info
        if isinstance(source.data, list):
            source.text = source.data[0]
            source.data = source.data[1].data

        result: ParseResult = super().parse(source)

        if not result.resolution_str is None and result.resolution_str:
            if not result.resolution_str.strip().endswith('%'):
                result.resolution_str = result.resolution_str.strip() + '%'

        result.data = source.text
        result.text = original

        return result



from typing import Dict, Pattern, List

from recognizers_text.utilities import RegExpUtility
from recognizers_text.culture import Culture
from recognizers_text.parser import ParseResult
from recognizers_number.culture import CultureInfo
from recognizers_number.number.parsers import NumberParserConfiguration
from recognizers_number.resources.english_numeric import EnglishNumeric


class EnglishNumberParserConfiguration(NumberParserConfiguration):
    @property
    def cardinal_number_map(self) -> Dict[str, int]:
        return self._cardinal_number_map

    @property
    def ordinal_number_map(self) -> Dict[str, int]:
        return self._ordinal_number_map

    @property
    def round_number_map(self) -> Dict[str, int]:
        return self._round_number_map

    @property
    def culture_info(self):
        return self._culture_info

    @property
    def digital_number_regex(self) -> Pattern:
        return self._digital_number_regex

    @property
    def fraction_marker_token(self) -> str:
        return self._fraction_marker_token

    @property
    def negative_number_sign_regex(self) -> Pattern:
        return self._negative_number_sign_regex

    @property
    def half_a_dozen_regex(self) -> Pattern:
        return self._half_a_dozen_regex

    @property
    def half_a_dozen_text(self) -> str:
        return self._half_a_dozen_text

    @property
    def lang_marker(self) -> str:
        return self._lang_marker

    @property
    def non_decimal_separator_char(self) -> str:
        return self._non_decimal_separator_char

    @property
    def decimal_separator_char(self) -> str:
        return self._decimal_separator_char

    @property
    def word_separator_token(self) -> str:
        return self._word_separator_token

    @property
    def written_decimal_separator_texts(self) -> List[str]:
        return self._written_decimal_separator_texts

    @property
    def written_group_separator_texts(self) -> List[str]:
        return self._written_group_separator_texts

    @property
    def written_integer_separator_texts(self) -> List[str]:
        return self._written_integer_separator_texts

    @property
    def written_fraction_separator_texts(self) -> List[str]:
        return self._written_fraction_separator_texts

    def __init__(self, culture_info=None):
        if culture_info is None:
            culture_info = CultureInfo(Culture.English)

        self._culture_info = culture_info
        self._lang_marker = EnglishNumeric.LangMarker
        self._decimal_separator_char = EnglishNumeric.DecimalSeparatorChar
        self._fraction_marker_token = EnglishNumeric.FractionMarkerToken
        self._non_decimal_separator_char = EnglishNumeric.NonDecimalSeparatorChar
        self._half_a_dozen_text = EnglishNumeric.HalfADozenText
        self._word_separator_token = EnglishNumeric.WordSeparatorToken

        self._written_decimal_separator_texts = EnglishNumeric.WrittenDecimalSeparatorTexts
        self._written_group_separator_texts = EnglishNumeric.WrittenGroupSeparatorTexts
        self._written_integer_separator_texts = EnglishNumeric.WrittenIntegerSeparatorTexts
        self._written_fraction_separator_texts = EnglishNumeric.WrittenFractionSeparatorTexts

        self._cardinal_number_map = EnglishNumeric.CardinalNumberMap
        self._ordinal_number_map = EnglishNumeric.OrdinalNumberMap
        self._round_number_map = EnglishNumeric.RoundNumberMap
        self._negative_number_sign_regex = RegExpUtility.get_safe_reg_exp(
            EnglishNumeric.NegativeNumberSignRegex)
        self._half_a_dozen_regex = RegExpUtility.get_safe_reg_exp(
            EnglishNumeric.HalfADozenRegex)
        self._digital_number_regex = RegExpUtility.get_safe_reg_exp(
            EnglishNumeric.DigitalNumberRegex)

    def normalize_token_set(self, tokens: List[str], context: ParseResult) -> List[str]:
        frac_words: List[str] = list()
        tokens_len = len(tokens)
        i = 0
        while i < tokens_len:
            if '-' in tokens[i]:
                splited_tokens = tokens[i].split('-')
                if len(splited_tokens) == 2 and splited_tokens[1] in self.ordinal_number_map:
                    frac_words.append(splited_tokens[0])
                    frac_words.append(splited_tokens[1])
                else:
                    frac_words.append(tokens[i])
            elif i < tokens_len - 2 and tokens[i + 1] == '-':
                if tokens[i + 2] in self.ordinal_number_map:
                    frac_words.append(tokens[i])
                    frac_words.append(tokens[i + 2])
                else:
                    frac_words.append(
                        tokens[i] + tokens[i + 1] + tokens[i + 2])
                i += 2
            else:
                frac_words.append(tokens[i])
            i += 1

        return frac_words

    def resolve_composite_number(self, number_str: str) -> int:
        if '-' in number_str:
            numbers = number_str.split('-')
            ret = 0
            for num in numbers:
                if num in self.ordinal_number_map:
                    ret += self.ordinal_number_map[num]
                elif num in self.cardinal_number_map:
                    ret += self.cardinal_number_map[num]
            return ret
        elif number_str in self.ordinal_number_map:
            return self.ordinal_number_map[number_str]
        elif number_str in self.cardinal_number_map:
            return self.cardinal_number_map[number_str]

        return 0
